#include "application.h"
#include "acutil_audio.h"
#include "scenemgr.h"
#include "sprite.h"
#include "acgfx_graphics.h"
#include "acutil_input.h"
#include "acutil_config.h"
#include "gamemgr.h"
#include "acframe_appwindow.h"
#include <string>
#include "scriptmgr.h"
#include "acutil_filemgr.h"
#include "console.h"
#include "guimgr.h"
#include "guielement.h"
#include "acutil_log.h"
#include "acutil_timer.h"
#include "loader.h"
#include "physicsmgr.h"
#include "pathmgr.h"
#include "acutil_profiler.h"
#include "recorder.h"

using namespace std;
using namespace acFramework;
using namespace acGraphics;
using namespace acUtility;

CGameApplication *app = 0;

namespace acFramework
{
CApplication *CreateApplication()
{
	return new CGameApplication();
}
}

CGameApplication::CGameApplication() : CApplication()
{
	// Set global parameter
	app = this;

	sceneMgr    = 0;
	gameMgr     = 0;
	scriptMgr   = 0;
	console     = 0;
	guiMgr      = 0;
	physicsMgr  = 0;
	pathMgr     = 0;
	recorder    = 0;

	drawPhysics = false;
	recording   = false;
	playback    = false;

	successfulInit = false;
}

CGameApplication::~CGameApplication()
{
	// Clear global parameter
	app = 0;
}

int CGameApplication::Init(CConfig &cfg)
{
	// Determine the user's language
	// TODO: This should be moved to a standard System object
	// TODO: Use this info to set the in-game language
	{
		std::string	lang = "English";
		LANGID languageId = GetUserDefaultUILanguage();
		WORD primary = PRIMARYLANGID(languageId);
		switch( primary )
		{
			case LANG_CHINESE:		lang = "Chinese"; break;
			case LANG_JAPANESE:		lang = "Japanese"; break;
			case LANG_KOREAN:		lang = "Korean"; break;
			case LANG_ENGLISH:		lang = "English"; break;
			case LANG_GERMAN:		lang = "German"; break;
			case LANG_FRENCH:		lang = "French"; break;
			case LANG_SPANISH:		lang = "Spanish"; break;
			case LANG_ITALIAN:		lang = "Italian"; break;
			case LANG_SWEDISH:		lang = "Swedish"; break;
			case LANG_DUTCH:		lang = "Dutch"; break;
			case LANG_PORTUGUESE:	lang = "Portuguese"; break;
			default:				lang = "unknown";
		}

		LOG(("The language used by Windows is %s", lang.c_str()));
	}

	int r;
	bool anyErrors = false;
	wnd->UpdateWindowText("ProjectX");
	wnd->SetCustomCursor(true);
	
	// Set the default search paths for files

	// Initialize the console, so that the rest of the initialization can write to it if needed
	console = new CConsole();
	r = console->Init();
	if( r < 0 )
		return r;

	r = InitOptional(cfg);
	if( r < 0 )
	{
		console->WriteLine("Initialization failed...");
		LOG(("Initialization failed. Displaying console only."));

		// Abort initialization, but allow the app to run anyway
		return 0;
	}

	successfulInit = true;
	return 0;
}

int CGameApplication::InitOptional(CConfig &cfg)
{
	int r;

	// Retrieve application options
	cfg.GetAttrAsString("screenshotFolder", screenshotFolder, 0, "screens");

	// Instanciate the scene manager
	sceneMgr = new CSceneMgr();
	r = sceneMgr->Init(gfx);
	if( r < 0 )
	{
		console->WriteLine("Failed to initialize scene manager");
		return r;
	}
	sceneMgr->SetFixedViewport(true);

	// Instanciate the physics manager
	physicsMgr = new CPhysicsMgr();
	r = physicsMgr->Init();
	if( r < 0 )
	{
		console->WriteLine("Failed to initialize physics manager");
		return r;
	}

	// Instanciate the game manager
	gameMgr = new CGameMgr();
	r = gameMgr->Init(cfg);
	if( r < 0 )
	{
		console->WriteLine("Failed to initialize game manager");
		return r;
	}

	// Instanciate the path manager
	pathMgr = new CPathMgr();

	// Instanciate the gui manager
	guiMgr = new CGuiMgr();
	r = guiMgr->Init(cfg);
	if( r < 0 )
	{
		console->WriteLine("Failed to initialize gui manager");
		return r;
	}

	// Instanciate the script manager
	scriptMgr = new CScriptMgr();
	r = scriptMgr->Init();
	if( r < 0 )
	{
		console->WriteLine("Failed to initialize script manager");
		return r;
	}

	// Load key mappings
	for( int n = 0; n < cfg.GetAttrCount("keyAction"); n++ )
	{
		string str;
		cfg.GetAttrAsString("keyAction", str, n);
		int key, action;
		sscanf_s(str.c_str(), "%d,%d", &key, &action);
		input->MapKeyToAction(key, action);
	}

	// Load mouse mappings
	for( int n = 0; n < cfg.GetAttrCount("mouseAction"); n++ )
	{
		string str;
		cfg.GetAttrAsString("mouseAction", str, n);
		int mb, action;
		sscanf_s(str.c_str(), "%d,%d", &mb, &action);
		input->MapMouseButtonToAction(mb, action);
	}

	// Load the menu
	// TODO: Should the menu script be configurable?
	r = scriptMgr->LoadMenuScript("menu.as");
	if( r < 0 )
	{
		console->WriteLine("Failed to load menu");
		return r;
	}

	// Create the recorder
	recording = IsCmdLineArgSet("record");
	playback = IsCmdLineArgSet("play");
	if( recording || playback )
	{
		recorder = new CRecorder();
		if( recorder )
			if( recording )
				recorder->RecordSetup();
			else
				recorder->PlaybackSetup();
	}

	scriptMgr->CallMenuOnShowMenu(0);

	remainingTime = 0;

	return 0;
}

void CGameApplication::InitPlay()
{
	timer->Stop();

	// Refresh everything
	Uninit();

	CConfig cfg;
	cfg.LoadConfigFile(fileMgr->GetPathForFile("default.cfg").c_str());
	Init(cfg);

	timer->Start();
}

void CGameApplication::Uninit()
{
	successfulInit = false;

	if( recorder )
	{
		delete recorder;
		recorder = 0;
	}

	if( input )
	{
		input->ClearActionMap();
	}

	if( gameMgr )
	{
		delete gameMgr;
		gameMgr = 0;
	}

	if( pathMgr )
	{
		delete pathMgr;
		pathMgr = 0;
	}

	if( physicsMgr )
	{
		delete physicsMgr;
		physicsMgr = 0;
	}

	if( console )
	{
		delete console;
		console = 0;
	}

	if( guiMgr )
	{
		delete guiMgr;
		guiMgr = 0;
	}

	if( sceneMgr )
	{
		delete sceneMgr;
		sceneMgr = 0;
	}

	if( scriptMgr )
	{
		delete scriptMgr;
		scriptMgr = 0;
	}
}

void CGameApplication::Step()
{
	TimeIt("CGameApplication::Step");

	timer->NewFrame();

	remainingTime += timer->GetDeltaTime();

	// TODO: Smooth the delta over multiple frames to minimize stuttering
	if( remainingTime < 1/60.0 )
	{
		// Do at least one step
		remainingTime = 1/60.0f;
	}

	while( remainingTime >= 1/60.0f )
	{
		remainingTime -= 1/60.0f;

		// Get new user input
		input->PollDevices();

		// Record the user input
		if( recorder )
			if( recording )
				recorder->RecordFrame();
			else
				recorder->PlaybackFrame();

		// Let the console take care of input if it is open
		if( console->Frame() )
			break;
		
		// Process the rest of the application if the console is not open
		if( acWAS_PRESSED(input->GetKeyState(acKEY_F5)) )
		{
			// TODO: menu: If in menu, this should reload the menu, if in game it should reload the level
			LOG(("Reinitializing at user request"));

			InitPlay();
			return;
		}

		// Process ALT+RETURN
		if( acWAS_PRESSED(input->GetKeyState(acKEY_RETURN)) && acIS_PRESSED(input->GetKeyState(acKEY_LALT)) )
			wnd->SetFullscreen(!wnd->IsFullscreen());

		if( successfulInit )
		{
			// TODO: Should probably have a StartFrame method on the script manager for this
			scriptMgr->numAllocationsLastFrame = scriptMgr->numAllocations;

			{
				TimeIt("CGameApplication::Frame::GameMgr");
				// Only call frame on the gameMgr if the game is running
				if( gameMgr->isGameOn && !gameMgr->isInMenu )
					gameMgr->Frame();
			}

			{
				TimeIt("CGameApplication::Frame::GuiMgr");
				// Always call Frame on the gui
				guiMgr->Frame();
			}

			scriptMgr->GarbageCollect();
		}

		// TODO: Should be possible to configure escape key
		if( acWAS_PRESSED(input->GetKeyState(acKEY_ESCAPE)) )
			gameMgr->PauseGame();

		// If the user pressed print screen we must exit the loop so the 
		// scene can be rendered and the screenshot saved to disk
		if( acWAS_PRESSED(input->GetKeyState(acKEY_SYSRQ)) )
			break;
	}

	// Render the scene
	Render();
}

void CGameApplication::Quit()
{
	::PostQuitMessage(0);
}

void CGameApplication::Render()
{
	TimeIt("CGameApplication::CustomRender");

	if( gfx == 0 )
		return;

	gfx->BeginScene();

	if( successfulInit )
	{
		// The scene and physics is not drawn if in menu mode
		if( !gameMgr->isInMenu )
		{
			{
				TimeIt("CGameApplication::Render::SceneMgr");
				sceneMgr->Render();
			}

			if( drawPhysics )
				physicsMgr->DebugDraw();

			// TODO: Add flag for this, and allow it to be set by the console
			//pathMgr->DebugDraw();

			{
				TimeIt("CGameApplication::Render::GuiMgr");
				guiMgr->Render();
			}
		}
		else
		{
			// At least clear the screen
			gfx->SetRenderTarget(0, 0);
			gfx->Clear(CLEAR_TARGET|CLEAR_ZBUFFER, 0, 1, 0);
			
			// GUI should always be rendered, either it is the menu or the in-game gui that will be shown
			guiMgr->Render();
		}
	}
	else
	{
		// At least clear the screen
		gfx->SetRenderTarget(0, 0);
		gfx->Clear(CLEAR_TARGET|CLEAR_ZBUFFER, 0, 1, 0);
	}

	// The console is always assumed to have initialized
	console->Render();

	gfx->EndScene();

	// TODO: It should be possible to configure the key to use
	//       I should reserve an action id for print screen
	if( acWAS_PRESSED(input->GetKeyState(acKEY_SYSRQ)) )
	{
		// TODO: Should create the screenshot folder if it is not already there

		// TODO: Should avoid overwriting other screenshots by dynamically increasing filenumber

		// TODO: The screenshot should be cropped to the client rect

		gfx->SaveScreenshot((screenshotFolder + "/screen.png").c_str());
	}

	// Update the screen
	gfx->Present();
}