#include "console.h"
#include "acgfx_font.h"
#include "acutil_filemgr.h"
#include "application.h"
#include "acgfx_graphics.h"
#include "acutil_timer.h"
#include "scenemgr.h"
#include "scriptmgr.h"
#include "acutil_input.h"
#include "acutil_unicode.h"
#include "gamemgr.h"
#include "gameobjectlink.h"

using namespace acGraphics;
using namespace acUtility;
using namespace std;

CConsole::CConsole()
{
	font = 0;
	lines.resize(100);
	currentLine = 0;
	visibleLines = 0;
	takeInput = false;
	showStats = false;
	showFps   = false;
}

CConsole::~CConsole()
{
	Uninit();
}

int CConsole::Init()
{
	font = new CFont();
	int r = font->Init(app->fileMgr->GetPathForFile("console.fnt").c_str(), 
		               app->fileMgr->GetPathForFile("font.fx").c_str(), 
			           app->render);

	font->SetTextEncoding(UTF8);

	return r;
}

void CConsole::Uninit()
{
	if( font )
	{
		delete font;
	}
}

bool CConsole::Frame()
{
	if( !takeInput )
	{
		visibleLines -= 1/60.0f;
		if( visibleLines < 0 ) visibleLines = 0;

		// TODO: Should use a reserved action id to allow configuration of key
		if( acWAS_PRESSED(app->input->GetKeyState(acKEY_GRAVE)) )
		{
			app->timer->Stop();
			Show();
		}
	}
	else
	{
		// TODO: Allow repeating commands by pressing the up arrow
		for( int n = 0; n < 256; n++ )
		{
			if( acWAS_PRESSED(app->input->GetKeyState(n)) )
			{
				// TODO: Cursor up/down should show previous commands
				// TODO: Page up/down should scroll console

				if( n == acKEY_RETURN || n == acKEY_ESCAPE )
					break;

				if( n == acKEY_BACK )
				{
					if( inputString.length() > 0 )
					{
						// Need to remove the entire unicode char
						while( (inputString[inputString.length()-1] & 0xC0) == 0x80 )
							inputString.resize(inputString.length() - 1);

						inputString.resize(inputString.length() - 1);
					}
				}
				else
				{
					// Determine the character
					unsigned short ch[4];
					int c = app->input->KeyToUnicode(n, ch, 4);
					unsigned char *ch2 = (unsigned char *)ch;
					while( c-- > 0 )
					{
						// Convert UTF16 to UTF8
						unsigned int len;
						int val = DecodeUTF16(ch2, &len);
						ch2 += len;
						if( val >= 0 )
						{
							unsigned char buf[5];
							EncodeUTF8(val, buf, &len);
							if( len > 0 )
								inputString.append((char*)buf, len);
						}
					}
				}
			}
		}

		if( acWAS_PRESSED(app->input->GetKeyState(acKEY_RETURN)) || acWAS_PRESSED(app->input->GetKeyState(acKEY_ESCAPE)) )
		{
			// Close the console first so that any output from the command is still shown
			Hide();

			if( acWAS_PRESSED(app->input->GetKeyState(acKEY_RETURN)) )
			{
				if( ExecuteCommand(inputString) )
					return false;
			}

			inputString = "";
			app->timer->Start();
		}
	}

	// Tell the application if we hold the input or not
	return takeInput;
}

void CConsole::Render()
{
	DisplayStatistics();

	DisplayLines();
}

void CConsole::DisplayStatistics()
{
	char buf[256];

	int w, h;
	app->gfx->GetScreenDimensions(&w, &h);

	font->PrepareEffect();
	font->PreparePixelPerfectOutput();

	if( app->successfulInit )
	{
		if( showFps )
		{
			// TODO: Draw the font with a drop shadow to improve visibility
			sprintf_s(buf, 256, "fps: %0.2f", app->timer->GetFps());
			font->Write(0, h-1*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);
		}

		if( showStats )
		{
			UINT gcObjects, gcDestroyed, gcDetected;
			gcObjects = app->scriptMgr->GetGCObjects(&gcDestroyed, &gcDetected);
			font->Write(0, h-4*font->GetHeight(), 0, "SCRIPT SYSTEM", 0, FONT_ALIGN_LEFT);
			sprintf_s(buf, 256, "gc objects: %d", gcObjects);
			font->Write(0, h-5*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);
			sprintf_s(buf, 256, "gc destroyed: %d", gcDestroyed);
			font->Write(0, h-6*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);
			sprintf_s(buf, 256, "gc detected: %d", gcDetected);
			font->Write(0, h-7*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);
			sprintf_s(buf, 256, "total mem allocated: %d", app->scriptMgr->totalAllocatedMemory);
			font->Write(0, h-8*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);
			sprintf_s(buf, 256, "total num allocations: %d", app->scriptMgr->numAllocations);
			font->Write(0, h-9*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);
			sprintf_s(buf, 256, "current allocations: %d", app->scriptMgr->numAllocations - app->scriptMgr->numFrees);
			font->Write(0, h-10*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);
			sprintf_s(buf, 256, "allocations in frame: %d", app->scriptMgr->numAllocations - app->scriptMgr->numAllocationsLastFrame);
			font->Write(0, h-11*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);

			font->Write(0, h-13*font->GetHeight(), 0, "GAME SYSTEM", 0, FONT_ALIGN_LEFT);
			sprintf_s(buf, 256, "game objects: %d", app->gameMgr->objects.size());
			font->Write(0, h-14*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);

			font->Write(0, h-16*font->GetHeight(), 0, "SCENE", 0, FONT_ALIGN_LEFT);
			int numDrawCalls, numMaterialChanges;
			app->sceneMgr->GetStats(&numDrawCalls, &numMaterialChanges);
			sprintf_s(buf, 256, "draw calls: %d", numDrawCalls);
			font->Write(0, h-17*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);
			sprintf_s(buf, 256, "material changes: %d", numMaterialChanges);
			font->Write(0, h-18*font->GetHeight(), 0, buf, 0, FONT_ALIGN_LEFT);
		}
	}
}

void CConsole::DisplayLines()
{
	int w, h;
	app->gfx->GetScreenDimensions(&w, &h);

	font->PrepareEffect();
	font->PreparePixelPerfectOutput();

	int offset = takeInput ? 0 : 1;

	for( int n = 1; n <= int(visibleLines+0.999f); n++ )
	{
		int line = currentLine - n;
		if( line < 0 ) 
			line += int(lines.size());

		// TODO: Fade the last line with time
		font->Write(0, (n-offset)*font->GetHeight() - font->GetBottomOffset(), 0, lines[line].c_str(), 0, FONT_ALIGN_LEFT);
	}

	if( takeInput )
	{
		// TODO: Should blink the caret
		font->Write(0, -font->GetBottomOffset(), 0, ("> " + inputString + "_").c_str(), 0, FONT_ALIGN_LEFT);
	}
}

void CConsole::WriteLine(const char *text)
{
	lines[currentLine++] = text;
	if( currentLine >= int(lines.size()) )
		currentLine = 0;
	visibleLines++;
	if( visibleLines > 20 )
		visibleLines = 20;
}

void CConsole::WriteMultipleLines(const char *text)
{
	int len = int(strlen(text));
	while( text )
	{
		string str;
		const char *ch = strchr(text, '\n');
		if( ch )
		{
			str.assign(text, ch - text);
			text = ch + 1;
		}
		else
		{
			str.assign(text);
			text = 0;
		}

		if( text || str.length() )
			WriteLine(str.c_str());
	}
}

void CConsole::Show()
{
	// Show as many lines as possible
	visibleLines = 20;
	takeInput = true;
}

void CConsole::Hide()
{
	visibleLines = 0;
	takeInput = false;
}

bool CConsole::IsTakingInput()
{
	return takeInput;
}

// Returns true if the console is destroyed with the execution
bool CConsole::ExecuteCommand(string &cmd)
{
	if( cmd == "" ) return false;

	WriteLine(cmd.c_str());

	string args;
	int n = cmd.find(" ");
	if( n != cmd.npos )
	{
		args = cmd.substr(n+1);
		cmd = cmd.substr(0,n);
	}

	if( cmd == "debug" )
	{
		if( args == "off" )
		{
			showStats = false;
			app->drawPhysics = false;
		}
		// TODO: Should be debug script, debug particles, etc
		else if( args == "stats" )
		{
			showStats = true;
		}
		else if( args == "physics" )
		{
			app->drawPhysics = true;
		}
	}
	else if( cmd == "show" && args == "fps" )
	{
		showFps = true;
	}
	else if( cmd == "hide" && args == "fps" )
	{
		showFps = false;
	}
	else if( cmd == "help" )
	{
		WriteLine("- debug (off|stats|physics)");
		WriteLine("- show (fps)");
		WriteLine("- hide (fps)");
		WriteLine("- play");
		WriteLine("- time (multiplier)");
	}
	else if( cmd == "time" )
	{
		sscanf_s(args.c_str(), "%f", &app->gameMgr->timeMultiplier);
		if( app->gameMgr->timeMultiplier < 0.01f )
			app->gameMgr->timeMultiplier = 0.01f;
		else if( app->gameMgr->timeMultiplier > 100 )
			app->gameMgr->timeMultiplier = 100;
	}
	else if( cmd == "play" )
	{
		app->InitPlay();
		return true;
	}
	else
	{
		WriteLine("Unknown command!");
	}

	return false;
}