#include <assert.h>
#include <vector>

#include "gamemgr.h"
#include "gameobjfactory.h"
#include "gameobject.h"
#include "gameobjectlink.h"
#include "acgfx_graphics.h"
#include "application.h"
#include "scenemgr.h"
#include "scriptmgr.h"
#include "guimgr.h"
#include "physicsmgr.h"
#include "physicsobject.h"
#include "acutil_config.h"
#include "loader.h"
#include "generator.h"
#include "acutil_filemgr.h"
#include "console.h"
#include "sprite.h"
#include "gamegroup.h"
#include "acframe_appwindow.h"
#include "pathmgr.h"

#ifdef SendMessage
// Remove the macro set by windows.h
#undef SendMessage
#endif

using namespace acGraphics;
using namespace acUtility;
using namespace std;

CGameMgr::CGameMgr()
{
	timeMultiplier = 1.0f;
	isGameOn = false;
	isInMenu = true;
}

CGameMgr::~CGameMgr()
{
	for( unsigned int n = 0; n < objectFactories.size(); n++ )
		if( objectFactories[n] )
			delete objectFactories[n];
	
	for( unsigned int n = 0; n < objects.size(); n++ )
		if( objects[n] )
			delete objects[n];

	for( unsigned int n = 0; n < groups.size(); n++ )
		if( groups[n] )
			delete groups[n];
}

int CGameMgr::Init(CConfig &cfg)
{
	cfg.GetAttrAsFloat("timeMultiplier", timeMultiplier, 0, 1);

	cfg.GetAttrAsString("gameScript", gameScript);

	return 0;
}

unsigned int CGameMgr::GetRandomSeed()
{
	return randomizer.getSeed();
}

void CGameMgr::SetRandomSeed(unsigned int seed)
{
	randomizer.setSeed(seed);
}

void CGameMgr::Frame()
{
	if( isInMenu )
	{
		// The game is paused, don't do anything
		frameDeltaTime = 0;
		return;
	}

	// Keep the original count, so that objects spawned by others don't get to update this frame
	unsigned int count = (unsigned int)objects.size();

	// TODO: Perhaps the physics simulation should be updated in smaller steps
	//       if so the collision callback must be executed in each step, and objects
	//       that are killed or spawned should be updated, but all other updates
	//       should only occur in the larger step
	app->physicsMgr->Frame();

	// Call each object's Frame method
	for( unsigned int n = 0; n < count; n++ )
		objects[n]->Frame();

	// Invoke the touch messages
	count = (unsigned int)objectsToTouch.size();
	for( unsigned int n = 0; n < count; n += 2 )
	{
		if( objectsToTouch[n]->obj &&
			objectsToTouch[n]->obj->m_scriptObj )
		{
			app->scriptMgr->CallEntityOnTouch(objectsToTouch[n]->obj->m_factory->scriptType, objectsToTouch[n]->obj->m_scriptObj, objectsToTouch[n+1]);
		}
		objectsToTouch[n]->Release();
		objectsToTouch[n+1]->Release();
	}
	objectsToTouch.erase(objectsToTouch.begin(), objectsToTouch.begin()+count);

	// Send the messages to the objects
	count = (unsigned int)objectMessages.size();
	for( unsigned int n = 0; n < count; n++ )
	{
		if( objectMessages[n].to->obj && 
			objectMessages[n].to->obj->m_scriptObj )
		{
			app->scriptMgr->CallEntityOnMessage(objectMessages[n].to->obj->m_factory->scriptType, objectMessages[n].to->obj->m_scriptObj, objectMessages[n].msgId, objectMessages[n].arg);
		}
		objectMessages[n].to->Release();
	}
	objectMessages.erase(objectMessages.begin(), objectMessages.begin()+count);

	// Synchronize the scene and physics objects
	count = (unsigned int)objects.size();
	for( unsigned int n = 0; n < count; n++ )
		objects[n]->SynchSceneAndPhysics();

	for( unsigned int n = 0; n < groups.size(); n++ )
		groups[n]->SynchSceneAndPhysics();

	// Check for collisions
	for( int n = 0; n < app->physicsMgr->GetCollisionCount(); n++ )
	{
		CPhysicsObject *a, *b;
		vector2 collisionPoint, collisionNormal;
		float appliedImpulse;
		app->physicsMgr->GetCollision(n, &a, &b, &collisionPoint, &collisionNormal, &appliedImpulse);

		CGameObjectLink *goa = a ? a->m_link : 0;
		CGameObjectLink *gob = b ? b->m_link : 0;

		if( goa == 0 && a && a->m_group && a->m_group )
			goa = a->m_group->anchor;
		if( gob == 0 && b && b->m_group && b->m_group )
			gob = b->m_group->anchor;

		if( goa && goa->obj )
			goa->obj->OnCollide(gob, collisionPoint, collisionNormal, appliedImpulse);
		if( gob && gob->obj )
			gob->obj->OnCollide(goa, collisionPoint, -collisionNormal, appliedImpulse);
	}

	// Check for sensor proximities
	for( int n = 0; n < app->physicsMgr->GetSensorProximityCount(); n++ )
	{
		CPhysicsObject *sensor, *object;
		bool isNearing;
		app->physicsMgr->GetSensorProximity(n, &sensor, &object, &isNearing);

		CGameObject *goSensor = sensor ? sensor->m_link->obj : 0;
		CGameObject *goObject = object ? object->m_link->obj : 0;

		if( goSensor && goObject )
			goSensor->OnProximity(goObject->m_link, isNearing);
	}

	// The objects that move out of the world are killed
	for( unsigned int n = 0; n < objects.size(); n++ )
	{
		vector2 pos = objects[n]->GetPosition();
		if( pos.x > 1000 || pos.x < -1000 || pos.y > 1000 || pos.y < -1000 )
			QueueObjectKill(objects[n]->m_link);
	}

	// At the end, when all processing have completed, destroy the objects that were killed during the frame
	for( unsigned int n = 0; n < objectsToKill.size(); n++ )
	{
		CGameObjectLink *link = objectsToKill[n];
		if( link->obj )
			DestroyObject(link->obj);
		link->Release();
	}
	objectsToKill.resize(0);

	// Was the level ended?
	if( callEndLevel )
		app->scriptMgr->CallGameOnEndLevel(endLevelCode);
}

void CGameMgr::SetTimeMultiplier(float multiplier)
{
	timeMultiplier = multiplier;
}

void CGameMgr::SetCameraPosition(const vector2 &pos)
{
	app->sceneMgr->SetCameraPosition(pos);
}

vector2 CGameMgr::ScreenToWorld(const acUtility::vector2 &pos)
{
	// First map the input from normalized vector to screen resolution
	vector2 pos2 = app->guiMgr->NormalizePosition(pos);

	// Convert the normalized position to window (screen)
	int x, y;
	app->wnd->GetSize(x, y);

	x = int(x*pos2.x);
	y = y - int(y*pos2.y);

	// Let the scene manager convert to the world coordinate according to camera position
	return app->sceneMgr->ConvertScreenToWorld(x, y);
}

vector2 CGameMgr::GetCameraPosition()
{
	return app->sceneMgr->GetCameraPosition();
}

float CGameMgr::GetFrameDeltaTime()
{
	return 1/60.0f * timeMultiplier;
}

float CGameMgr::GetRandomNumber()
{
	return randomizer.getNumber();
}

CGameObject *CGameMgr::CreateObject(const char *type, const vector2 &pos)
{
	// Find the factory for this object type
	CGameObjectFactory *factory = 0;
	for( unsigned int n = 0; n < objectFactories.size(); n++ )
	{
		if( objectFactories[n]->typeName == type )
		{
			factory = objectFactories[n];
			break;
		}
	}

	// Create the factory if it wasn't found
	if( factory == 0 )
	{
		factory = new CGameObjectFactory();
		factory->Load(type);
		objectFactories.push_back(factory);
	}

	// Create the object
	CGameObject *obj = factory->CreateObject(pos);
	
	// Add the object to the game, then return it
	objects.push_back(obj);

	return obj;
}

void CGameMgr::DestroyObject(CGameObject *obj)
{
	for( unsigned int n = 0; n < objects.size(); n++ )
	{
		if( objects[n] == obj )
		{
			objects.erase(objects.begin()+n);
			break;
		}
	}
	delete obj;
}

// TODO: Maybe this should be made into a generic message queue
void CGameMgr::QueueObjectKill(CGameObjectLink *link)
{
	// Put it on the queue to be killed at the end of the frame
	objectsToKill.push_back(link);
	link->AddRef();
}

void CGameMgr::QueueTouch(CGameObjectLink *origin, const CGameObjectLink *other)
{
	if( !origin || !other ) return;

	// Put the pair on the queue
	// It's safe to remove the constness. It was only there because the  
	// scripts aren't supposed to directly manipulate other game objects
	objectsToTouch.push_back(const_cast<CGameObjectLink*>(other));
	objectsToTouch.push_back(origin);
	other->AddRef();
	origin->AddRef();
}

void CGameMgr::SendMessage(const CGameObjectLink *to, unsigned int message, CScriptHandle arg)
{
	if( !to ) return;

	// TODO: Should figure out who is sending the message (when it is from a game object)
	//       The origin should be informed to the destination

	// Put the message on the queue
	// It's safe to remove the constness. It was only there because the
	// scripts aren't supposed to directly manipulate other game objects
	SObjMsg msg;
	msg.to    = const_cast<CGameObjectLink*>(to); // The reference is already accounted for
	msg.msgId = message;
	msg.arg   = arg;                              // The script handle will take care of the reference counting
	objectMessages.push_back(msg);
}


void CGameMgr::AddGroup(CGameGroup *grp)
{
	groups.push_back(grp);
}

CGameObject *CGameMgr::AddObject(const std::string &type, const vector2 &pos, int zOrder, const std::vector<std::string> &nameValuePairs, CGameGroup *group)
{
	CGameObject *obj = CreateObject(type.c_str(), pos);
	if( obj->m_sprite )
	{
		// TODO: This should be done by CreateObject
		obj->m_sprite->zOrder = zOrder;
	}

	// Set the properties
	for( unsigned int n = 0; n < nameValuePairs.size(); n += 2 )
	{
		if( nameValuePairs[n] == "name" )
			obj->m_name = nameValuePairs[n+1];
		else 
		{
			// TODO: This should be done by script mgr
			// Find a suitable property in the script controller object
			if( obj->m_scriptObj )
			{
				for( unsigned int p = 0; p < obj->m_scriptObj->GetPropertyCount(); p++ )
				{
					if( nameValuePairs[n] == obj->m_scriptObj->GetPropertyName(p) )
					{
						// Check the type of the property
						// TODO: Must support other types too
						int typeId = obj->m_scriptObj->GetPropertyTypeId(p);
						if( app->scriptMgr->IsEntityTypeId(typeId) )
						{
							// The value is a name that needs to be matched with the object
							// TODO: What if the other object hasn't been created yet
							//       Create a list of deferred attributions, when a new object with a name
							//       is added, then add this one. At the end when the loader signals that
							//       no more objects are to be added the game manager should warn on unfound links
							CGameObjectLink *link = FindObjectByName(nameValuePairs[n+1]);
							// FindObjectByName already increases the refCount
							*(CGameObjectLink**)obj->m_scriptObj->GetAddressOfProperty(p) = link;
						}
						else if( typeId == asTYPEID_FLOAT )
						{
							float value = float(atof(nameValuePairs[n+1].c_str()));
							*(float*)obj->m_scriptObj->GetAddressOfProperty(p) = value;
						}
						else if( typeId == app->scriptMgr->vector2TypeId )
						{
							vector2 value;
							sscanf_s(nameValuePairs[n+1].c_str(), "%f,%f", &value.x, &value.y);
							*(vector2*)obj->m_scriptObj->GetAddressOfProperty(p) = value;
						}
					}
				}
			}
		}
	}

	if( group )
	{
		if( obj->m_factory->groupAnchor )
		{
			// This object is an anchor for a group of sprites and blockers
			obj->m_group = group;

			// The group must also know of the object
			group->SetAnchor(obj->m_link);
		}
		else if( obj->m_factory->weight == 0 )
		{
			// TODO: Use a specific attribute that better defines this

			// This is a static object and should move together with the group
			group->AddObject(obj->m_link);
		}
	}

	// TODO: Call onInit after setting all properties; Only when the loader signals that everything is loaded

	return obj;
}

CGameObjectLink *CGameMgr::SpawnObject(const std::string &type, const vector2 &pos, const vector2 &velocity)
{
	CGameObject *obj = CreateObject(type.c_str(), pos);

	obj->SetVelocity(velocity);

	// TODO: Call onInit after setting all properties
	
	// Return a link to the object
	obj->m_link->AddRef();
	return obj->m_link;
}

// TODO: Should support continued calls, so that all objects can be enumerated
CGameObjectLink *CGameMgr::FindObjectByName(const std::string &name)
{
	for( int n = 0; n < int(objects.size()); n++ )
	{
		if( objects[n] && objects[n]->m_name == name )
		{
			objects[n]->m_link->AddRef();
			return objects[n]->m_link;
		}
	}

	return 0;
}

vector2 CGameMgr::GetNearestPathNode(const vector2 &from, const vector2 &to)
{
	return app->pathMgr->GetNearestPathNode(from, to);
}

// TODO: Might be interesting to allow the menu script to pass an argument to the game script
//       This can for example be used to start a random skirmish, instead of starting a new campaign.
int CGameMgr::StartGame()
{
	// Clear the GUI
	app->guiMgr->RemoveAll();

	// Load the game script
	int r = app->scriptMgr->LoadGameScript(gameScript.c_str());
	if( r < 0 )
	{
		app->console->WriteLine("Failed to load game script");
		return r;
	}

	// Call the start game script function
	app->scriptMgr->CallGameOnStartGame();

	// Set condition that the game is running
	isGameOn = true;
	isInMenu = false;

	return 0;
}

void CGameMgr::SetGravity(const acUtility::vector2 &grav)
{
	app->physicsMgr->SetGlobalGravity(grav);
}

void CGameMgr::RestartLevel()
{
	assert( isGameOn );

	// Clear the GUI
	app->guiMgr->RemoveAll();

	// Show the in-game GUI
	app->scriptMgr->CallGuiOnShowGui();

	// Call the game' onRestartLevel handler
	app->scriptMgr->CallGameOnRestartLevel();

	isInMenu = false;
}

void CGameMgr::ResumeGame()
{
	assert( isGameOn );

	// Clear the GUI
	app->guiMgr->RemoveAll();

	// Show the in-game GUI
	app->scriptMgr->CallGuiOnShowGui();

	// Set condition that the game is running
	isInMenu = false;
}

void CGameMgr::QuitGame()
{
	app->Quit();
}

bool CGameMgr::IsGameOn()
{
	return isGameOn;
}

void CGameMgr::EndGame()
{
	// Set condition that the game is not running anymore
	isGameOn = false;

	ShowMenu(0);
}

void CGameMgr::PauseGame()
{
	ShowMenu(0);
}

void CGameMgr::ShowMenu(unsigned int id)
{
	// Set the condition that the game is paused
	isInMenu = true;

	app->guiMgr->RemoveAll();

	app->scriptMgr->CallMenuOnShowMenu(id);
}

void CGameMgr::InitLevel()
{
	// The level is no longer completed
	callEndLevel = false;

	// Clean everything in the sceneMgr, physicsMgr, etc
	for( unsigned int n = 0; n < objects.size(); n++ )
		delete objects[n];
	objects.clear();
	for( unsigned int n = 0; n < groups.size(); n++ )
		delete groups[n];
	groups.clear();
	app->sceneMgr->ClearScene();
	app->physicsMgr->ClearWorld();

	// Erase the level properties
	levelProperties.clear();
}

int CGameMgr::LoadLevel(const string &level)
{
	InitLevel();

	// Load the new level
	CLoader loader;
	int r = loader.Load(app, app->sceneMgr, this, app->fileMgr->GetPathForFile(level.c_str()).c_str());
	if( r < 0 )
	{
		app->console->WriteLine(("Failed to load map file " + level).c_str());
		return r;
	}

	return 0;
}

int CGameMgr::GenerateLevel(const string &logic)
{
	InitLevel();

	// Generate the new level
	CGenerator generator;
	int r = generator.Generate(logic);
	if( r < 0 )
	{
		app->console->WriteLine(("Failed to generate map with logic " + logic).c_str());
		return r;
	}

	return 0;
}

void CGameMgr::EndLevel(int code)
{
	// Set internal flag so the script event is called at the end of the frame
	callEndLevel = true;
	endLevelCode = code;
}

float CGameMgr::RayCast(const CGameObjectLink *ignoreObj, int query, const vector2 &start, const vector2 &end, const CGameObjectLink **firstObj, vector2 *hitVelocity)
{
	const CGameObject *obj = ignoreObj ? ignoreObj->obj : 0;
	const CPhysicsObject *phys = obj ? obj->m_physicsObject : 0;
	CPhysicsObject *firstContact = 0;

	float timeOfImpact = app->physicsMgr->RayCast(phys, query, start, end, &firstContact);

	if( firstObj && firstContact && firstContact->m_link )
	{
		*firstObj = firstContact->m_link;
		(*firstObj)->AddRef();
	}

	if( hitVelocity )
	{
		if( firstContact )
			*hitVelocity = firstContact->GetVelocity();
		else
			*hitVelocity = vector2(0,0);
	}
				
	if( ignoreObj )
		ignoreObj->Release();

	return timeOfImpact;
}

bool CGameMgr::GetProperty(const string &prop, __int64 &out)
{
	// First look in the game script
	if( app->scriptMgr->GetPropertyFromGameScript(prop, out) )
		return true;

	// Look in the level created properties
	map<string, __int64>::iterator it;
	it = levelProperties.find(prop);
	if( it != levelProperties.end() )
	{
		out = it->second;
		return true;
	}

	out = 0;
	return false;
}

bool CGameMgr::SetProperty(const string &prop, __int64 val)
{
	// First look in the game script
	if( app->scriptMgr->SetPropertyFromGameScript(prop, val) )
		return true;

	// Set a property for the level
	map<string, __int64>::iterator it;
	it = levelProperties.find(prop);
	if( it != levelProperties.end() )
		it->second = val;
	else
		levelProperties.insert(map<string, __int64>::value_type(prop, val));

	return true;
}

void CGameMgr::FindObjectsWithinRange(const vector2 &center, float radius, vector<CGameObject*> &foundObjects)
{
	// TODO: Should optimize finding objects by some hierarchical groupings
	// TODO: Allow filtering objects with a bitmask, e.g. only search for enemies, or only search for pickups, etc
	// TODO: Allow filtering objects by visibility, e.g. perform raytracing to determine if an object is visible or not

	// Loop through the existing objects and add the ones that are close enough to the vector
	
	// Multiply the radius so we don't have to do a lot of squareroots
	radius *= radius; 
	for( unsigned int n = 0; n < objects.size(); n++ )
	{
		vector2 pos = objects[n]->GetPosition();
		if( (pos - center).lengthSq() < radius )
			foundObjects.push_back(objects[n]);
	}
}
