#include "gameobject.h"
#include "sprite.h" 
#include "application.h"
#include "scriptmgr.h"
#include "gamemgr.h"
#include "gameobjectlink.h"
#include "acutil_log.h"
#include "scenemgr.h"
#include "spriteanimation.h"
#include "gameobjfactory.h"
#include "physicsobject.h"
#include "physicsmgr.h"
#include "acutil_audio.h"
#include "acutil_filemgr.h"
#include "console.h"
#include "gamegroup.h"

using namespace acUtility;
using namespace std;

CGameObject::CGameObject(CGameObjectFactory *factory)
{
	m_factory         = factory;
	m_link            = new CGameObjectLink(this);
	m_sprite          = 0;
	m_scriptObj       = 0;
	m_animation       = 0;
	m_physicsObject   = 0;
	m_awake           = true; // TODO: Should be possible to set initially asleep
	m_group           = 0;
	m_moveSpeed       = 0;
}

CGameObject::~CGameObject()
{
	if( m_physicsObject )
	{
		app->physicsMgr->DestroyObject(m_physicsObject);
		m_physicsObject = 0;
	}

	if( m_sprite )
	{
		// Remove the scene object from the scene
		app->sceneMgr->RemoveSprite(m_sprite);
		delete m_sprite;
		m_sprite = 0;
	}

	if( m_link )
	{
		// Sever the link to this object, then release the link
		m_link->obj = 0;
		m_link->Release();
		m_link = 0;
	}

	if( m_scriptObj )
	{
		m_scriptObj->Release();
		m_scriptObj = 0;
	}
}

void CGameObject::Kill()
{
	app->gameMgr->QueueObjectKill(m_link);
}

void CGameObject::Frame()
{
	// Animate the sprite
	// TODO: This should possibly be moved to the sceneMgr
	if( m_animation )
	{
		m_timeUntilNextFrame -= app->gameMgr->GetFrameDeltaTime();
		while( m_timeUntilNextFrame < 0 && m_animation )
		{
			m_currentFrame++;
			if( m_currentFrame >= m_animation->times.size() )
			{
				if( m_animation->nextAnimation.length() == 0 )
				{
					// The animation ended
					m_sprite->material = m_animation->images[m_animation->images.size()-1];
					m_sprite->offset   = m_animation->offsets[m_animation->images.size()-1];
					m_animation = 0;
					break;
				}
				else if( m_animation->nextAnimation == m_animation->name )
				{
					m_currentFrame = 0;
					m_timeUntilNextFrame += m_animation->times[m_currentFrame];
				}
				else 
				{
					m_currentFrame = 0;
					m_animation = m_factory->GetAnimation(m_animation->nextAnimation.c_str());
					if( m_animation )
						m_timeUntilNextFrame += m_animation->times[m_currentFrame];
				}
			}
			else
			{
				m_timeUntilNextFrame += m_animation->times[m_currentFrame];
			}
		}
		if( m_animation )
		{
			m_sprite->material = m_animation->images[m_currentFrame];
			m_sprite->offset   = m_animation->offsets[m_currentFrame];
		}
	}

	// Is there a controlled movement set?
	if( m_moveSpeed )
	{
		// If the actual speed is not the same as the set speed, then
		// we assume it was changed by a collision, and we stop the controlled movement
		if( fabs(GetVelocity().lengthSq() - m_moveSpeed*m_moveSpeed) > 0.0001f )
		{
			// Clear the moveSpeed. Note, this doesn't stop any uncontrolled movement that may be happening.
			// Let the script entity decide if it should stop the uncontrolled movement or not.
			m_moveSpeed = 0;

			// Wake up the controller
			m_awake = true;
		}
		else
		{
			// TODO: Should prevent overshooting already within the physicsmgr. Currently
			//       there may be collisions triggered that never really should happen

			// As the actual movement is performed by the physics manager before we get here,
			// we need to check for potential overshot, and if there is any move the object
			// back to the target position.
			vector2 vec = m_targetPos - GetPosition();
			if( dot(vec, GetVelocity()) < 0 )
			{
				// We've reached the target, so let's stop
				SetPosition(m_targetPos);
				SetVelocity(vector2(0,0));

				m_moveSpeed = 0;

				// Wake up the controller
				m_awake = true;
			}
		}
	}

	// Check timers
	// This is done before calling onThink to allow a timer to wake up the object
	// Get the count before, so we don't end up calling newly created timers with zero wait time
	// TODO: If the timer kills another, we might end up calling an incorrect timer anyway
	unsigned int count = m_timers.size();
	for( unsigned int i = 0; i < count; i++ )
	{
		m_timers[i].time -= app->gameMgr->GetFrameDeltaTime();
		if( m_timers[i].time <= 0 )
		{
			int id = m_timers[i].id;

			m_timers.erase(m_timers.begin()+i);
			i--; count--;

			if( m_scriptObj )
				app->scriptMgr->CallEntityOnTimer(m_factory->scriptType, m_scriptObj, id);
		}
	}

	// Call the onThink if the object is awake
	if( m_awake && m_scriptObj )
	{
		app->scriptMgr->CallEntityOnThink(m_factory->scriptType, m_scriptObj);
	}
}

// Returns a negative time if no timer is set
float CGameObject::GetTimer(int id) const
{
	// Find the smallest time for the id
	float smallest = -1;
	for( unsigned int i = 0; i < m_timers.size(); i++ )
	{
		if( m_timers[i].id == id )
		{
			if( smallest < 0 || smallest > m_timers[i].time )
				smallest = m_timers[i].time;
		}
	}

	return smallest;
}

// TODO: Perhaps add a flag for replacing an existing timer
void CGameObject::SetTimer(float time, int id)
{
	STimer timer = {time, id};
	m_timers.push_back(timer);
}

void CGameObject::KillTimer(int id)
{
	for( unsigned int i = 0; i < m_timers.size(); i++ )
	{
		if( m_timers[i].id == id )
		{
			m_timers.erase(m_timers.begin()+i);
			i--;
		}
	}
}

void CGameObject::SetAnimation(const string &anim)
{
	if( m_animation && m_animation->name == anim )
		return;

	m_animation = m_factory->GetAnimation(anim.c_str());
	m_currentFrame = 0;
	if( m_animation )
	{
		m_timeUntilNextFrame = m_animation->times[0];
		m_sprite->material = m_animation->images[0];
		m_sprite->offset = m_animation->offsets[0];
	}
	else
	{
		m_sprite->material = app->sceneMgr->GetMaterialIndex(m_factory->sprite.c_str());
		m_sprite->offset = vector2(0,0);
	}
}

void CGameObject::OnCollide(const CGameObjectLink *obj, const vector2 &v, const vector2 &n, float approachSpeed)
{
	// Wake up the object, even if there is no collision handler
	m_awake = true;

	// By default the m_moveSpeed is cleared, as the MoveTo command was interrupted
	if( m_moveSpeed != 0 )
		m_moveSpeed = 0;

	if( m_factory->collisionSound )
	{
		// TODO: If there is more than one sound effect, a random one should be selected
		// TODO: the sound effect to use can also be selected based on the material of the two objects
		// TODO: The collision sound should be based on the actual impact, i.e. it depends on the mass ratio between the two objects
		app->audio->PlaySound(m_factory->collisionSound, approachSpeed/100);
	}

	if( m_scriptObj )
		app->scriptMgr->CallEntityOnCollide(m_factory->scriptType, m_scriptObj, obj, v, n, approachSpeed);
}

void CGameObject::OnProximity(const CGameObjectLink *obj, bool isNearing)
{
	if( m_scriptObj )
		app->scriptMgr->CallEntityOnProximity(m_factory->scriptType, m_scriptObj, obj, isNearing);
}

void CGameObject::SetPosition(const vector2 &pos)
{
	if( m_sprite )
		m_sprite->pos = pos;

	if( m_physicsObject )
		m_physicsObject->SetPosition(pos);

	if( m_group )
		m_group->SetPosition(pos);
}

vector2 CGameObject::GetPosition() const
{
	if( m_physicsObject )
		return m_physicsObject->GetPosition();

	if( m_sprite )
		return m_sprite->pos;

	if( m_group )
		return m_group->GetPosition();

	return vector2(0,0);
}

void CGameObject::SetRotation(float rot)
{
	if( m_sprite )
		m_sprite->rotation = rot;

	if( m_physicsObject )
		m_physicsObject->SetRotation(rot);

	// TODO: The group's rotation also needs to be updated
}

float CGameObject::GetRotation() const
{
	if( m_sprite )
		return m_sprite->rotation;
	
	if( m_physicsObject )
		return m_physicsObject->GetRotation();

	return 0;
}

void CGameObject::SynchSceneAndPhysics()
{
	if( m_sprite && m_physicsObject )
	{
		m_sprite->pos      = m_physicsObject->GetPosition();
		m_sprite->rotation = m_physicsObject->GetRotation();
	}
}

void CGameObject::SetVelocity(const vector2 &v)
{
	// No more controlled movement
	m_moveSpeed = 0;

	if( m_physicsObject )
		m_physicsObject->SetVelocity(v);

	if( m_group )
		m_group->SetVelocity(v);
}

vector2 CGameObject::GetVelocity() const
{
	if( m_physicsObject )
		return m_physicsObject->GetVelocity();

	if( m_group )
		return m_group->GetVelocity();

	return vector2(0,0);
}

bool CGameObject::GetProperty(const std::string &prop, __int64 &out)
{
	if( m_scriptObj == 0 ) return false;

	for( unsigned int n = 0; n < m_scriptObj->GetPropertyCount(); n++ )
	{
		if( prop == m_scriptObj->GetPropertyName(n) )
		{
			void *ptr = m_scriptObj->GetAddressOfProperty(n);
			int valueTypeId = m_scriptObj->GetPropertyTypeId(n);

			return app->scriptMgr->ConvertTypeToInt64(ptr, valueTypeId, out);
		}
	}

	return false;
}

void CGameObject::MoveTo(const vector2 &pos, float speed)
{
	if( speed < 0 )
		speed = 0;		

	if( speed > 0 )
	{
		vector2 vec = pos - GetPosition();
		float len = vec.length();
		if( len > 0.0001f )
		{
			vec *= speed/vec.length();
			SetVelocity(vec);
		}
		else
		{
			SetVelocity(vector2(0,0));
			speed = 0;
		}
	}

	m_targetPos = pos;
	m_moveSpeed = speed;
}

void CGameObject::ClearGravity()
{
	if( m_physicsObject )
		m_physicsObject->ClearCustomGravity();
}

void CGameObject::SetGravity(const vector2 &g)
{
	if( m_physicsObject )
		m_physicsObject->SetCustomGravity(g);
}

void CGameObject::SetCollision(bool collision)
{
	if( m_physicsObject )
		m_physicsObject->SetCollision(collision);
}

const string &CGameObject::GetType() const
{
	return m_factory->typeName;
}

void CGameObject::Touch(const CGameObjectLink *other)
{
	app->gameMgr->QueueTouch(m_link, other);
}

void CGameObject::SetTransparency(float fraction)
{
	if( fraction < 0 ) fraction = 0;
	if( fraction > 1 ) fraction = 1;
	if( m_sprite )
		m_sprite->opaque = 1 - fraction;
}

void CGameObject::PlaySound(const string &snd, float volume)
{
	// Postfix the extension
	// TODO: The extension .sfx can also be used. This would be a text file similar to
	//       the material, where properties can be defined, e.g. a different volume, or
	//       a different pitch, etc.
	string sndFile = snd+".wav";

	// TODO: Should use shared sound buffers, to avoid multiple loads (needs CSoundMgr)
	//       GameObjectFactory should keep a cache of used sounds, so 
	//       these can be released when the object type is no longer in use
	//       The game factory should also allow pre-loading sounds by listing
	//       them in the object file. This will avoid staggering during game play
	CSoundBuffer *sound;
	app->audio->CreateSoundBufferFromFile(app->fileMgr->GetPathForFile(sndFile.c_str()).c_str(), &sound);
	if( sound )
	{
		// TODO: Should allow setting the volume from script
		// TODO: Should play the sound according to the entity's relative position with the camera
		app->audio->PlaySound(sound, volume);
		sound->Release();
	}
}

