#include "gameobjfactory.h"
#include "gameobject.h"
#include "acgfx_graphics.h"
#include "sprite.h"
#include "scenemgr.h"
#include "acutil_config.h"
#include "acutil_log.h"
#include "application.h"
#include "scriptmgr.h"
#include "acutil_filemgr.h"
#include "spriteanimation.h"
#include "physicsmgr.h"
#include "physicsobject.h"
#include "gameobjectlink.h"
#include "console.h"
#include "acutil_audio.h"

using namespace acGraphics;
using namespace acUtility;
using namespace acFramework;
using namespace std;

CGameObjectFactory::CGameObjectFactory()
{
	scriptType = 0;
	collisionSound = 0;
	offset = vector2(0,0);
}

CGameObjectFactory::~CGameObjectFactory()
{
	for( unsigned int n = 0; n < animations.size(); n++ )
		if( animations[n] )
			delete animations[n];

	if( collisionSound )
		collisionSound->Release();
}

int CGameObjectFactory::Load(const char *type)
{
	typeName = type;

	// Load the object description file
	string filename = typeName + ".obj";
	CConfig file;
	int r = file.LoadConfigFile(app->fileMgr->GetPathForFile(filename.c_str()).c_str());
	if( r < 0 ) return r;

	// Load the sprite name
	file.GetAttrAsString("sprite", sprite);

	// Load the size
	string sizeStr;
	file.GetAttrAsString("size", sizeStr);
	if( sizeStr != "" )
	{
		sscanf_s(sizeStr.c_str(), "%f,%f", &size.x, &size.y);
	}

	// Sprite offset
	string spriteOffset;
	file.GetAttrAsString("offset", spriteOffset);
	if( spriteOffset != "" )
	{
		sscanf_s(spriteOffset.c_str(), "%f,%f", &offset.x, &offset.y);
	}

	file.GetAttrAsBool("groupAnchor", groupAnchor, 0);

	// Set collision properties
	string collisionShape;
	file.GetAttrAsString("collisionShape", collisionShape);
	int n = collisionShape.find_first_of(",");
	string name = collisionShape.substr(0,n);
	if( name == "circle" )
	{
		collisionFixedRotation = false;
		collisionShapeType = 1;
		sscanf_s(collisionShape.c_str()+n+1, "%f", &collisionHalfSize.x);
	}
	else if( name == "box" )
	{
		collisionFixedRotation = false;
		collisionShapeType = 2;
		sscanf_s(collisionShape.c_str()+n+1, "%f,%f", &collisionHalfSize.x, &collisionHalfSize.y);
	}
	else if( name == "capsule" )
	{
		collisionFixedRotation = true;
		collisionShapeType = 3;
		sscanf_s(collisionShape.c_str()+n+1, "%f,%f", &collisionHalfSize.x, &collisionHalfSize.y);
	}
	else if( name == "bullet" )
	{
		collisionShapeType = 4;
	}
	else
	{
		collisionFixedRotation = false;
		collisionShapeType = 0;
		collisionHalfSize = vector2(0,0);
	}

	file.GetAttrAsBool("fixedRotation", collisionFixedRotation, 0, collisionFixedRotation);

	file.GetAttrAsFloat("weight", weight, 0, 1.0f);

	string collProp;
	file.GetAttrAsString("collisionProperty", collProp, 0, "solid");
	if( collProp == "passthrough" )
		collisionProperty = PP_PASSTHROUGH;
	else if( collProp == "sensor" )
		collisionProperty = PP_SENSOR;
	else
		collisionProperty = PP_SOLID;


	string collSound;
	file.GetAttrAsString("collisionSound", collSound);
	if( collSound.length() )
		app->audio->CreateSoundBufferFromFile(app->fileMgr->GetPathForFile(collSound.c_str()).c_str(), &collisionSound);

	// Load the controller script
	string scriptFile;
	if( file.GetAttrAsString("script", scriptFile) )
	{
		// Tell the script manager to load and build the controller script
		scriptType = app->scriptMgr->LoadControllerScript(scriptFile.c_str());
	}

	// Load the animations
	for( int n = 0; n < file.GetAttrCount("animation"); n++ )
	{
		string tmp;
		file.GetAttrAsString("animation", tmp, n);

		int i = tmp.find_first_of(",");

		// TODO: If there is no next, the animation simply stops at the last frame
		CSpriteAnimation *anim = new CSpriteAnimation();
		animations.push_back(anim);
		anim->name          = tmp.substr(0,i);
		anim->nextAnimation = tmp.substr(i+1);
		
		// Load the frames
		for( int m = 0; m < file.GetAttrCount(anim->name.c_str()); m++ )
		{
			file.GetAttrAsString(anim->name.c_str(), tmp, m);

			i = tmp.find_first_of(",");

			anim->images.push_back(app->sceneMgr->GetMaterialIndex(tmp.substr(0,i).c_str()));

			float time, xOff = 0, yOff = 0;
			sscanf_s(tmp.substr(i+1).c_str(), "%f,%f,%f", &time, &xOff, &yOff);

			anim->times.push_back(time);
			anim->offsets.push_back(vector2(xOff,yOff));
		}

		if( anim->images.size() == 0 )
		{
			string msg = "Error in animation ";
			msg += type;
			msg += " ";
			msg += anim->name;
			app->console->WriteLine(msg.c_str());
			LOG((msg.c_str()));

			animations.pop_back();
			delete anim;
		}
	}

	return 0;
}

CGameObject *CGameObjectFactory::CreateObject(const vector2 &pos)
{
	CGameObject *obj = new CGameObject(this);
	
	// Create the scene object and add it to the scene
	if( sprite != "" )
	{
		obj->m_sprite = new CSprite();
		obj->m_sprite->material = app->sceneMgr->GetMaterialIndex(sprite.c_str());
		obj->m_sprite->size     = size;
		obj->m_sprite->offset   = offset;
		app->sceneMgr->AddSprite(obj->m_sprite);
	}

	// Set collision shape
	if( collisionShapeType )
	{
		// TODO: Should be possible to have a different offset for the collisionShape
		if( collisionShapeType == 1 )
			obj->m_physicsObject = app->physicsMgr->CreateObjectAsCircle(pos, collisionHalfSize.x, weight, collisionProperty);
		else if( collisionShapeType == 2 )
			obj->m_physicsObject = app->physicsMgr->CreateObjectAsBox(pos, collisionHalfSize.x, collisionHalfSize.y, weight, collisionProperty);
		else if( collisionShapeType == 3 )
			obj->m_physicsObject = app->physicsMgr->CreateObjectAsCapsule(pos, collisionHalfSize.x, collisionHalfSize.y, weight, collisionProperty);
		else if( collisionShapeType == 4 )
			obj->m_physicsObject = app->physicsMgr->CreateObjectAsBullet(pos, weight, collisionProperty);
		obj->m_physicsObject->SetFixedRotation(collisionFixedRotation);
		obj->m_physicsObject->m_link = obj->m_link;
		obj->m_link->AddRef();
	}

	// Create the script object instance that controls this object
	if( scriptType )
	{
		obj->m_scriptObj = app->scriptMgr->CreateScriptObject(scriptType, obj->m_link);
	}

	obj->SetPosition(pos);

	return obj;
}

CSpriteAnimation *CGameObjectFactory::GetAnimation(const char *anim)
{
	for( unsigned int n = 0; n < animations.size(); n++ )
	{
		if( animations[n]->name == anim )
			return animations[n];
	}

	return 0;
}