#include "generator.h"
#include "sprite.h"
#include "scenemgr.h"
#include "application.h"
#include "physicsmgr.h"
#include "gamemgr.h"
#include "scriptmgr.h"
#include "pathmgr.h"
#include "acutil_math.h"

using namespace std;
using namespace acGraphics;
using namespace acUtility;


CGenerator::CGenerator()
{
	grid = 0;
}

CGenerator::~CGenerator()
{
	if( grid )
		delete[] grid;
}

int CGenerator::Generate(const string &script)
{
	// TODO: Should be possible to specificy seed, so a specific random level can be regenerated
	rand.setSeed(unsigned int(app->gameMgr->GetRandomNumber()*1000000));

	// The logic for generating the map is implemented in the script
	int r = app->scriptMgr->CallGeneratorScript(script.c_str(), this);

	// Finalize will create the scene, set up the physics world, initiate the game objects, etc
	Finalize();

	return r;
}

void CGenerator::Finalize()
{
	// Don't do anything if the grid hasn't been created
	if( grid == 0 ) return;

	// Get the materials for the tile set
	// TODO: These could be defined by the script
	CSceneMgr *sceneMgr = app->sceneMgr;
	unsigned int floor = sceneMgr->GetMaterialIndex("floor.png");
	unsigned int wall = sceneMgr->GetMaterialIndex("wall.png");

	// Create the tiles for the scene based on the grid layout
	for( unsigned int y = 0; y < height; y++ )
	{
		for( unsigned int x = 0; x < width; x++ )
		{
			char value = grid[x + y*width];

			// Don't care about values that are not 0 or 1 at the moment
			if( value == 0 || value == 1 )
			{
				CSprite *sprite = new CSprite;
				if( value == 0 )
					sprite->material = wall;
				else
					sprite->material = floor;
				sprite->pos = vector2(float(x)+0.5f,float(y)+0.5f);
				sprite->size = vector2(1,1);
				sprite->zOrder = 100;
				sceneMgr->AddSprite(sprite);
			}
		}
	}

	// Create the physics blockers
	CPhysicsMgr *physMgr = app->physicsMgr;
	for( unsigned int y = 0; y < height; y++ )
	{
		for( unsigned int x = 0; x < width; x++ )
		{
			// Look for transitions wall to floor
			// TODO: Merge collision boxes to have fewer
			if( grid[x + y*width] == 0 )
			{
				if( At(x-1, y) == 1 || At(x+1, y) == 1 ||
					At(x, y-1) == 1 || At(x, y+1) == 1 )
				{
					physMgr->AddCollisionBox(vector2(float(x), float(y)), vector2(float(x+1), float(y+1)), false);
				}
			}
		}
	}

	// Initialize the path manager
	CPathMgr *pathMgr = app->pathMgr;
	pathMgr->SetArea(width, height);
	for( unsigned int y = 0; y < height; y++ )
		for( unsigned int x = 0; x < width; x++ )
			if( grid[x + y*width] == 1 )
				pathMgr->SetWalkable(x, y, true);
	pathMgr->Optimize();
}

void CGenerator::SetMapSize(unsigned int width, unsigned int height)
{
	// Release old grid
	if( grid ) 
		delete[] grid;

	// Allocate the grid size
	grid = new char[width*height];
	this->width  = width;
	this->height = height;

	// Fill the grid with walls
	memset(grid, 0, width*height);
}

void CGenerator::FillRect(unsigned int x, unsigned int y, unsigned int w, unsigned int h, char value)
{
	if( grid == 0 ) return;

	// Limit the area to the available grid
	if( y+h > height ) { if( y >= height ) h = 0; else h = height - y; }
	if( x+w > width  ) { if( x >= width  ) w = 0; else w = width  - x; }

	// Fill the rect
	for( unsigned int dy = 0; dy < h; dy++ )
	{
		char *d = grid + (dy+y)*width + x;
		for( unsigned int dx = 0; dx < w; dx++, d++ )
			*d = value;
	}
}

char &CGenerator::At(unsigned int x, unsigned int y)
{
	// If the coordinate is outside the grid, return a dummy 
	static char outside = 0;
	if( grid == 0 ) return outside;
	if( x >= width ) return outside;
	if( y >= height ) return outside;

	// Return a reference to the grid element
	return grid[y*width + x];
}

void CGenerator::AddObject(const string &type, float x, float y, int zOrder, const vector<string> &nameValuePairs)
{
	app->gameMgr->AddObject(type, vector2(x,y), zOrder, nameValuePairs, 0);
}

float CGenerator::GetRandomNbr()
{
	return rand.getNumber();
}