#include "guielement.h"
#include "acgfx_graphics.h"
#include "acgfx_dynrender.h"
#include "application.h"
#include "guimgr.h"
#include "angelscript.h"
#include "scriptmgr.h"
#include "acgfx_font.h"

using namespace acGraphics;
using namespace acUtility;

CGuiElement::CGuiElement(const vector2 &pos, const vector2 &size)
{
	refCount      = 1;

	position      = pos;
	this->size    = size;
	center        = vector2(0,0);
	rotation      = 0;
	scale         = vector2(1,1);
	zPosition     = 0;
	srcImageIndex = -1;
	srcBottomLeft = vector2(0,0);
	srcTopRight   = vector2(1,1);
	fontIndex     = app->guiMgr->defaultFontIndex;
	parent        = 0;
	align         = 1;
	color         = 0xFFFFFFFF;

	onClickHandler     = 0;
	onMouseOverHandler = 0;
	onTimerHandler     = 0;
}

CGuiElement::~CGuiElement()
{
	// Release all children
	for( int n = 0; n < (int)children.size(); n++ )
	{
		children[n]->parent = 0;
		children[n]->Release();
	}
	children.resize(0);

	if( onClickHandler )
		onClickHandler->Release();
	if( onMouseOverHandler )
		onMouseOverHandler->Release();
	if( onTimerHandler )
		onTimerHandler->Release();
}

void CGuiElement::AddRef()
{
	refCount++;
}

void CGuiElement::Release()
{
	if( --refCount == 0 )
		delete this;
}

void CGuiElement::AddChild(CGuiElement *child)
{
	if( child->parent ) return;

	// TODO: Don't allow the addition if this window is a child or grand child of the new child itself

	children.push_back(child);
	child->AddRef();

	// Add a weak reference to the parent
	child->parent = this;
}

void CGuiElement::RemoveChild(CGuiElement *child)
{
	if( child->parent != this ) return;

	for( unsigned int n = 0; n < children.size(); n++ )
	{
		if( children[n] == child )
		{
			children[n] = children.back();
			children.pop_back();
			child->parent = 0;
			child->Release();
			break;
		}
	}
}

void CGuiElement::Render(const matrix4x4 &parentMtx)
{
	// Set world matrix
	matrix4x4 mtx;
	mtx.setIdentity();
	mtx *= matrix4x4::translation(-center.x, -center.y, 0);
	mtx *= matrix4x4::scaling(scale.x, scale.y, 1);
	mtx *= matrix4x4::rotationZ(rotation);
	mtx *= matrix4x4::translation(position.x, position.y, 0);
	mtx *= parentMtx;
	
	CGraphics *gfx = app->gfx;
	gfx->SetMatrix(MTX_WORLD, mtx);

	// Draw the element
	DrawSelf();

	// Render all children recursively
	for( int n = 0; n < (int)children.size(); n++ )
	{
		children[n]->Render(mtx);
	}
}

void CGuiElement::DrawSelf()
{
	app->guiMgr->DrawRect(vector2(0,0), size, srcImageIndex, srcBottomLeft, srcTopRight);
	app->guiMgr->DrawText(vector2(0,0), text, fontIndex, align, color);

/*
	// TODO: Only draw this on debug
	app->gfx->SetTexture(0);
	app->render->Begin(RENDER_LINE_LIST);

	// Defined box
	app->render->VtxColor(0xFFFFFFFF);
	app->render->VtxPos(0,0,0);
	app->render->VtxPos(size.x, 0, 0);
	app->render->VtxPos(size.x, 0, 0);
	app->render->VtxPos(size.x, size.y, 0);
	app->render->VtxPos(size.x, size.y, 0);
	app->render->VtxPos(0, size.y, 0);
	app->render->VtxPos(0, size.y, 0);
	app->render->VtxPos(0, 0, 0);

	// Text area
	if( text != "" )
	{
		vector2 ll, tr;
		app->guiMgr->GetTextRect(ll, tr, text, fontIndex);
		app->render->VtxColor(0xFF00FF00);
		app->render->VtxPos(ll.x, ll.y,0);
		app->render->VtxPos(tr.x, ll.y, 0);
		app->render->VtxPos(tr.x, ll.y, 0);
		app->render->VtxPos(tr.x, tr.y, 0);
		app->render->VtxPos(tr.x, tr.y, 0);
		app->render->VtxPos(ll.x, tr.y, 0);
		app->render->VtxPos(ll.x, tr.y, 0);
		app->render->VtxPos(ll.x, ll.y, 0);
	}

	app->render->End();
*/
}

bool CGuiElement::IsMouseOver()
{
	vector2 pos = app->guiMgr->mousePos;

	// TODO: Should check children too (must also include parent transform)

	// Transform the mouse position into the local gui element space
	// TODO: Should do this with a 3x3 matrix instead
	matrix4x4 mtx;
	mtx.setIdentity();
	mtx *= matrix4x4::translation(-center.x, -center.y, 0);
	mtx *= matrix4x4::scaling(scale.x, scale.y, 1);
	mtx *= matrix4x4::rotationZ(rotation);
	mtx *= matrix4x4::translation(position.x, position.y, 0);

	mtx = matrix4x4::inverse(mtx);
	vector3 pos3 = vectorTransformCoord(vector3(pos.x, pos.y, 0), mtx);
	pos = vector2(pos3.x, pos3.y);

	// If there is no image and there is text, then use the text extension as hit rect
	vector2 ll, tr;
	if( text != "" && srcImageIndex == -1 )
	{
		app->guiMgr->GetTextRect(ll, tr, text, fontIndex);
		if( align == acGraphics::FONT_ALIGN_CENTER )
		{
			float w = 0.5f*(tr.x - ll.x);
			ll.x -= w;
			tr.x -= w;
		}
		else if( align == acGraphics::FONT_ALIGN_RIGHT )
		{
			float w = tr.x - ll.x;
			ll.x -= w;
			tr.x -= w;
		}
	}
	else
	{
		ll = vector2(0,0);
		tr = ll + size;
	}

	// Is the mouse position within the borders
	if( pos.x < ll.x || pos.x > tr.x ||
	    pos.y < ll.y || pos.y > tr.y )
		return false;

	return true;
}

bool CGuiElement::HandleMouseOver()
{
	if( onMouseOverHandler == 0 )
		return false;

	if( !IsMouseOver() )
		return false;

	if( app->guiMgr->currentMouseOver != this )
	{
		// Call the onMouseOver for the previous element to notify it that the mouse moved away
		// As the previous element may have been removed, we must find the real element, if it still exists
		CGuiElement *el = app->guiMgr->ElementExists(app->guiMgr->currentMouseOver);
		if( el && el->onMouseOverHandler ) 
			app->scriptMgr->CallGuiOnMouseOverHandler(el->onMouseOverHandler, false, el);

		// Call the onMouseOver to notify the element that the mouse moved over it
		app->scriptMgr->CallGuiOnMouseOverHandler(onMouseOverHandler, true, this);
		app->guiMgr->currentMouseOver = this;
	}

	return true;
}

bool CGuiElement::IsChild(CGuiElement *el)
{
	for( unsigned int n = 0; n < children.size(); n++ )
	{
		if( children[n] == el ) return true;
		if( children[n]->IsChild(el) ) return true;
	}

	return false;
}

bool CGuiElement::HandleClick()
{
	if( onClickHandler == 0 )
		return false;

	if( !IsMouseOver() )
		return false;

	// Call the onClick event handler
	app->scriptMgr->CallGuiOnClickHandler(onClickHandler, this);

	return true;
}

void CGuiElement::Frame()
{
	// The timers might actually cause the element to be deleted, so
	// we need to increase the reference temporarily to prevent a too
	// early destruction
	AddRef();

	// Check timers
	// Get the count before, so we don't end up calling newly created timers with zero wait time
	// TODO: If the timer kills another, we might end up calling an incorrect timer anyway
	unsigned int count = timers.size();
	for( unsigned int i = 0; i < count; i++ )
	{
		// TODO: Which timer to use should depend on the game mode
		timers[i].time -= 1/60.0f;
		if( timers[i].time <= 0 )
		{
			if( onTimerHandler )
				app->scriptMgr->CallGuiOnTimerHandler(onTimerHandler, timers[i].id, this);

			timers.erase(timers.begin()+i);
			i--; count--;
		}
	}

	Release();
}

void CGuiElement::SetTimer(float time, int id)
{
	timers.push_back(STimer(time,id));
}

void CGuiElement::KillTimer(int id)
{
	for( unsigned int n = 0; n < timers.size(); n++ )
	{
		if( timers[n].id == id )
		{
			timers.erase(timers.begin()+n);
			n--;
		}
	}
}