#include "guimgr.h"
#include "guielement.h"
#include "guilistbox.h"
#include "application.h"
#include "acgfx_graphics.h"
#include "acgfx_dynrender.h"
#include "acgfx_font.h"
#include "acutil_math.h"
#include "acutil_filemgr.h"
#include "acutil_input.h"
#include "acutil_log.h"
#include "scriptmgr.h"
#include <math.h>
#include <string>
#include "console.h"
#include "acutil_config.h"
#include "gamemgr.h"
#include "acframe_appwindow.h"

using namespace std;
using namespace acGraphics;
using namespace acUtility;

CGuiMgr::CGuiMgr()
{
	// TODO: Reference rect should be configurable by the script
	referenceWidth = 1024;
	referenceHeight = 768;
	mousePos = vector2(float(referenceWidth/2)+1, float(referenceHeight/2)+1);
	effectId = -1;
	hasScript = false;
	defaultFontIndex = -1;

	mouseElement = 0;
	currentMouseOver = 0;
}

CGuiMgr::~CGuiMgr()
{
	Uninit();
}

int CGuiMgr::Init(const CConfig &cfg)
{
	int r = 0;

	effectId = app->gfx->LoadEffect(app->fileMgr->GetPathForFile("gui.fx").c_str());
	if( effectId < 0 )
	{
		app->console->WriteMultipleLines(app->gfx->GetEffectCompileErrors());
		app->console->WriteLine("Failed to load gui.fx");
		return effectId;
	}

	// Get the default font and element texture from the file
	string defaultFontName;
	if( cfg.GetAttrAsString("guiDefaultFont", defaultFontName) )
	{
		defaultFontIndex = LoadFont(defaultFontName);
	}

	return r;
}

void CGuiMgr::Uninit()
{
	if( mouseElement )
	{
		mouseElement->Release();
		mouseElement = 0;
	}

	for( unsigned int n = 0; n < topLevelWindows.size(); n++ )
		topLevelWindows[n]->Release();
	topLevelWindows.resize(0);

	for( unsigned int n = 0; n < images.size(); n++ )
		if( images[n] )
			images[n]->Release();
	images.resize(0);

	for( unsigned int n = 0; n < fonts.size(); n++ )
		if( fonts[n] )
			delete fonts[n];
	fonts.resize(0);

	app->gfx->DropEffect(effectId);
}

void CGuiMgr::Frame()
{
	// Update mouse position

	// Get the Windows cursor position
	int x, y;
	app->input->GetCursorPos(&x, &y);
	mousePos.x = float(x)/1000000 * float(referenceWidth);
	mousePos.y = float(y)/1000000 * float(referenceHeight);

	if( mousePos.x < 0 ) mousePos.x = 0;
	if( mousePos.x > referenceWidth ) mousePos.x = float(referenceWidth);
	if( mousePos.y < 0 ) mousePos.y = 0;
	if( mousePos.y > referenceHeight ) mousePos.y = float(referenceHeight);

	if( mouseElement )
	{
		mouseElement->position = mousePos;

		// Look for mouse over events
		for( unsigned int n = topLevelWindows.size(); n-- > 0; )
		{
			if( topLevelWindows[n]->HandleMouseOver() )
				break;
		}

		// Make sure the mouse is still over the current element
		// As the previous element may have been removed, we must find the real element, if it still exists
		CGuiElement *el = ElementExists(currentMouseOver);
		if( el && !el->IsMouseOver() ) 
		{
			if( el->onMouseOverHandler )
				app->scriptMgr->CallGuiOnMouseOverHandler(el->onMouseOverHandler, false, el);
			currentMouseOver = false;
		}
	}

	if( app->gameMgr->isInMenu )
	{
		// Handle input
		if( acWAS_PRESSED(app->input->GetMouseButtonState(0)) )
		{
			// Find the gui element that was hit, so the event handler can be called
			for( unsigned int n = topLevelWindows.size(); n-- > 0; )
			{
				if( topLevelWindows[n]->HandleClick() )
					break;
			}
		}

		// Update the timers 
		// TODO: optimize: The gui elements should tell the gui manager when they have a 
		//                 timer, that way there won't be a need to traverse the whole tree
		for( unsigned int n = topLevelWindows.size(); n-- > 0; )
			topLevelWindows[n]->Frame();
	}
	else
	{
		// Call the script to update the in-game GUI
		if( hasScript )
			app->scriptMgr->CallGuiOnFrame();
	}
}

void CGuiMgr::RemoveAll()
{
	// Remove all existing elements
	for( unsigned int n = 0; n < topLevelWindows.size(); n++ )
		topLevelWindows[n]->Release();
	topLevelWindows.clear();
}

void CGuiMgr::Render()
{
	// Setup the view transform so that the gui is mapped to the full window
	CGraphics *gfx = app->gfx;

	gfx->SetEffect(effectId);

	matrix4x4 mtx;

	// Setup orthogonal view
	// Determine size of view port
	float minZ, maxZ;
	gfx->GetViewport(0,0,0,0,&minZ, &maxZ);

	// Origin is in lower-left corner
	mtx = matrix4x4::orthoOffCenterLH(0, (float)referenceWidth, 0, (float)referenceHeight, minZ, maxZ);
	gfx->SetMatrix(MTX_VIEW, mtx);

	mtx.setIdentity();
	gfx->SetMatrix(MTX_PROJ, mtx);

	// TODO: The sort should happen only when the zPosition of a window is modified.
	// Order the windows by z order. We use bubblesort here, because the windows 
	// will almost always be sorted already, thus a bubblesort is the fastest way of doing it.
	for( int n = 0; n < (int)topLevelWindows.size() - 1; n++ )
	{
		if( topLevelWindows[n]->zPosition < topLevelWindows[n+1]->zPosition )
		{	
			CGuiElement *swap = topLevelWindows[n+1];
			topLevelWindows[n+1] = topLevelWindows[n];
			topLevelWindows[n] = swap;

			// Move backwards
			for( unsigned int m = n; m > 0; m-- )
			{
				if( topLevelWindows[m]->zPosition <= topLevelWindows[m-1]->zPosition )
					break;

				topLevelWindows[m] = topLevelWindows[m-1];
				topLevelWindows[m-1] = swap;
			}
		}
	}

	// Render the windows, which are really just a bunch of rectangles with textures on them
	for( unsigned int n = 0; n < topLevelWindows.size(); n++ )
	{
		CGuiElement *el = topLevelWindows[n];

		matrix4x4 mtx = matrix4x4::identity();
		el->Render(mtx);
	}
}

CGuiElement *CGuiMgr::CreateElement(const vector2 &pos, const vector2 &size, CGuiElement *parent)
{
	CGuiElement *el = new CGuiElement(pos, size);
	if( parent )
		parent->AddChild(el);
	else
	{
		topLevelWindows.push_back(el);
		el->AddRef();
	}

	return el;
}

CGuiListBox *CGuiMgr::CreateListBox(const vector2 &pos, const vector2 &size, CGuiElement *parent)
{
	CGuiListBox *el = new CGuiListBox(pos, size);
	if( parent )
		parent->AddChild(el);
	else
	{
		topLevelWindows.push_back(el);
		el->AddRef();
	}

	return el;
}

void CGuiMgr::RemoveElement(CGuiElement *el)
{
	if( el->parent )
		el->parent->RemoveChild(el);
	else 
	{
		for( unsigned int n = 0; n < topLevelWindows.size(); n++ )
			if( topLevelWindows[n] == el )
			{
				topLevelWindows[n] = topLevelWindows.back();
				topLevelWindows.pop_back();
				el->Release();
				break;
			}
	}
}

CGuiElement *CGuiMgr::ElementExists(CGuiElement *el)
{
	// Traverse all elements to find the one with matching pointer
	for( unsigned int n = 0; n < topLevelWindows.size(); n++ )
	{
		if( topLevelWindows[n] == el ) return el;

		if( topLevelWindows[n]->IsChild(el) )
			return el;
	}

	return 0;
}

void CGuiMgr::SetMouseElement(CGuiElement *el)
{
	// Add the reference for the new element first, just in case it happens to be the same as we already have
	if( el ) el->AddRef();
	if( mouseElement ) mouseElement->Release();
	mouseElement = el;

	// Move it to the current mouse position
	mouseElement->position = mousePos;
}

int CGuiMgr::LoadImage(const std::string &name)
{
	// TODO: Should check if the image hasn't been loaded already
	CGraphicsBuffer *image = 0;
	string p = app->fileMgr->GetPathForFile(name.c_str());
	int r = app->gfx->CreateTextureFromFile(p.c_str(), &image);
	if( r < 0 )
		return r;
	images.push_back(image);
	return int(images.size())-1;
}

int CGuiMgr::LoadFont(const std::string &name)
{
	// TODO: Should check if the font hasn't been loaded already
	CFont *font = new CFont();
	int r = font->Init(app->fileMgr->GetPathForFile(name.c_str()).c_str(), 
		               app->fileMgr->GetPathForFile("font.fx").c_str(),
			           app->render);
	if( r < 0 )
	{
		LOG(("Failed to load font '%s' (%r)", name.c_str(), r));
		delete font;
		return -1;
	}

	fonts.push_back(font);
	return int(fonts.size())-1;
}

int CGuiMgr::LoadScript(const std::string &script)
{
	if( script == "" )
	{	
		hasScript = false;
		return 0;
	}

	hasScript = true;

	int r = app->scriptMgr->LoadGuiScript(script.c_str());
	if( r < 0 ) return r;

	// Call the Initialize() function
	app->scriptMgr->CallGuiOnShowGui();

	return r;
}

vector2 CGuiMgr::NormalizePosition(const vector2 &pos) const
{
	vector2 norm = pos;
	norm.x /= referenceWidth;
	norm.y /= referenceHeight;

	return norm;
}

vector2 CGuiMgr::GetNormalizedMousePos()
{
	return NormalizePosition(mousePos);
}

vector2 CGuiMgr::GetMousePos()
{
	return mousePos;
}

void CGuiMgr::DrawRect(const vector2 &bottomLeft, const vector2 &topRight, int srcImageIndex, const vector2 &srcBottomLeft, const vector2 &srcTopRight)
{
	if( srcImageIndex != -1 && srcImageIndex < int(images.size()) )
	{
		CGraphics *gfx = app->gfx;
		gfx->SetTexture(images[srcImageIndex], 0);

		// Draw the windows with the dynamic renderer
		app->render->Begin(RENDER_QUAD_LIST);
		app->render->VtxColor(0xFFFFFFFF);
		app->render->VtxTexCoord(srcBottomLeft.x, srcBottomLeft.y);
		app->render->VtxPos(bottomLeft.x, topRight.y, 0);
		app->render->VtxTexCoord(srcTopRight.x, srcBottomLeft.y);
		app->render->VtxPos(topRight.x, topRight.y, 0);
		app->render->VtxTexCoord(srcTopRight.x, srcTopRight.y);
		app->render->VtxPos(topRight.x, bottomLeft.y, 0);
		app->render->VtxTexCoord(srcBottomLeft.x, srcTopRight.y);
		app->render->VtxPos(bottomLeft.x, bottomLeft.y, 0);
		app->render->End();
	}
}

void CGuiMgr::DrawText(const acUtility::vector2 &pos, const std::string &text, int fontIndex, int align, unsigned int color)
{
	if( fontIndex >= 0 && fontIndex < int(fonts.size()) && text.length() > 0 )
	{
		CGraphics *gfx = app->gfx;
		matrix4x4 mtx = gfx->GetMatrix(MTX_WORLD);

		// TODO: Should cache the text by rendering to a dynamic texture. 
		//       That way it won't be necessary to switch effects here.

		// Switch to font effect
		fonts[fontIndex]->PrepareEffect();
		fonts[fontIndex]->SetColor(color);
		gfx->SetMatrix(MTX_WORLD, mtx);

		// Setup orthogonal view
		// Determine size of view port
		float minZ, maxZ;
		gfx->GetViewport(0,0,0,0,&minZ,&maxZ);

		matrix4x4 mtx2;
		mtx2 = matrix4x4::orthoOffCenterLH(0, (float)referenceWidth, 0, (float)referenceHeight, minZ, maxZ);
		gfx->SetMatrix(MTX_VIEW, mtx2);

		mtx2.setIdentity();
		gfx->SetMatrix(MTX_PROJ, mtx2);

		// TODO: Should also allow to align vertically.
		fonts[fontIndex]->WriteML(pos.x,pos.y,0,text.c_str(),0,align);

		// Switch to gui effect
		gfx->SetEffect(effectId);
		gfx->SetMatrix(MTX_WORLD, mtx);

		mtx2 = matrix4x4::orthoOffCenterLH(0, (float)referenceWidth, 0, (float)referenceHeight, minZ, maxZ);
		gfx->SetMatrix(MTX_VIEW, mtx2);

		mtx2.setIdentity();
		gfx->SetMatrix(MTX_PROJ, mtx2);
	}	
}

void CGuiMgr::GetTextRect(vector2 &ll, vector2 &tr, const string &text, int fontIndex)
{
	ll = tr = vector2(0,0);
	if( fontIndex >= 0 && fontIndex < int(fonts.size()) && text.length() > 0 )
	{
		tr.x = fonts[fontIndex]->GetTextWidth(text.c_str(), text.length(), &ll.y, &tr.y);
//		ll.y = fonts[fontIndex]->GetBottomOffset();
//		tr.y = fonts[fontIndex]->GetTopOffset();
	}
}

CFont *CGuiMgr::GetFont(int index)
{
	if( index >= 0 && index < (int)fonts.size() )
		return fonts[index];

	return 0;
}