#include "loader.h"
#include "sprite.h"
#include "scenemgr.h"
#include "application.h"
#include "acgfx_graphics.h"
#include "acutil_math.h"
#include <vector>
#include "acutil_log.h"
#include "gamemgr.h"
#include "acutil_config.h"
#include "gameobject.h"
#include "gameobjectlink.h"
#include "gamegroup.h"
#include "physicsmgr.h"
#include "physicsobject.h"

using namespace std;
using namespace acGraphics;
using namespace acUtility;

CLoader::CLoader()
{
}

CLoader::~CLoader()
{
}

int CLoader::Load(CGameApplication *app, CSceneMgr *sceneMgr, CGameMgr *gameMgr, const char *filename)
{
	CConfig file;
	int r = file.LoadConfigFile(filename);
	if( r < 0 )
	{
		LOG(("Failed to load map file '%s' (%d)", filename, r));
		return r;
	}

	map<int, CGameGroup *> groups;

	// Load the groups
	for( int n = 0; n < file.GetAttrCount("group"); n++ )
	{
		string args, name;
		file.GetAttrAsString("group", args, n);
		int groupId = atoi(args.c_str());
		int c = args.find_first_of(",");
		if( c != string::npos )
			name.assign(&args[c+1]);

		CGameGroup *group;
		map<int, CGameGroup*>::iterator it = groups.find(groupId);
		if( it != groups.end() )
			group = it->second;
		else
		{
			group = new CGameGroup();
			groups.insert(map<int,CGameGroup*>::value_type(groupId, group));
			gameMgr->AddGroup(group);
		}

		group->name = name;
	}

	// Process static scenery
	for( int n = 0; n < file.GetAttrCount("sprite"); n++ )
	{
		string args;
		file.GetAttrAsString("sprite", args, n);
		string type;
		float x,y,w,h;
		int zOrder;
		int c = (int)args.find_first_of(",");
		type.assign(args.c_str(), c);
		sscanf_s(&args[c+1], "%f,%f,%f,%f,%d", &x, &y, &w, &h, &zOrder);

		CSprite *sprite = new CSprite;
		sprite->material = sceneMgr->GetMaterialIndex(type.c_str());
		sprite->pos = vector2(x,y);
		sprite->size = vector2(w,h);
		sprite->zOrder = zOrder;
		sceneMgr->AddSprite(sprite);

		// Set the group for this sprite
		CGameGroup *group = 0;
		const char *grp = strstr(args.c_str(), ",g=");
		if( grp )
		{
			int groupId = atoi(grp+3);
			map<int, CGameGroup*>::iterator it = groups.find(groupId);
			if( it != groups.end() )
				group = it->second;
			else
			{
				group = new CGameGroup();
				groups.insert(map<int,CGameGroup*>::value_type(groupId, group));
				gameMgr->AddGroup(group);
			}

			if( group )
				group->AddSprite(sprite);
		}
	}

	// Process collision objects
	for( int n = 0; n < file.GetAttrCount("collision"); n++ )
	{
		string args;
		file.GetAttrAsString("collision", args, n);
		string type;
		vector2 v1,v2;
		int c = (int)args.find_first_of(",");
		type.assign(args.c_str(), c);
		sscanf_s(&args[c+1], "%f,%f,%f,%f", &v1.x, &v1.y, &v2.x, &v2.y);

		// Add the blocker to the correct group
		CGameGroup *group = 0;
		const char *grp = strstr(args.c_str(), ",g=");
		if( grp )
		{
			int groupId = atoi(grp+3);
			map<int, CGameGroup*>::iterator it = groups.find(groupId);
			if( it != groups.end() )
				group = it->second;
			else
			{
				group = new CGameGroup();
				groups.insert(map<int,CGameGroup*>::value_type(groupId, group));
				gameMgr->AddGroup(group);
			}
		}

		CPhysicsObject *obj = 0;
		if( type == "line" )
			obj = app->physicsMgr->AddCollisionLine(v1,v2, group ? true : false);
		else if( type == "box" )
			obj = app->physicsMgr->AddCollisionBox(v1,v2, group ? true : false);

		if( group )
		{
			group->AddBlocker(obj);
			obj->m_group = group;
		}
	}

	// Process the game objects
	for( int n = 0; n < file.GetAttrCount("object"); n++ )
	{
		string args;
		file.GetAttrAsString("object", args, n);
		string type;
		float x,y;
		int zOrder;
		int c = (int)args.find_first_of(",");
		type.assign(args.c_str(), c);
		sscanf_s(&args[c+1], "%f,%f,%d", &x, &y, &zOrder);

		CGameGroup *group = 0;

		// Scan the name value pairs
		c = args.find_first_of(",",c+1); 
		c = args.find_first_of(",",c+1);
		c = args.find_first_of(",",c+1);
		vector<string> nameValuePairs;
		while( c != string::npos )
		{
			int e = args.find_first_of("=", c+1);
			if( e == string::npos ) break;
			string value, name = args.substr(c+1, e - c-1);
			c = e;
			if( args[c+1] == '\'' )
			{
				e = args.find_first_of("'", c+2);
				value = args.substr(c+2, e - c-2);
				e++;
			}
			else 
			{
				e = args.find_first_of(",", c+1);
				value = args.substr(c+1, e - c-1);
			}

			if( name == "g" )
			{
				// Set the group for this object
				int groupId = atoi(value.c_str());
				map<int, CGameGroup*>::iterator it = groups.find(groupId);
				if( it != groups.end() )
					group = it->second;
				else
				{
					group = new CGameGroup();
					groups.insert(map<int,CGameGroup*>::value_type(groupId, group));
					gameMgr->AddGroup(group);
				}
			}
			else
			{
				nameValuePairs.push_back(name);
				nameValuePairs.push_back(value);
			}

			c = e;
		}

		CGameObject *obj = gameMgr->AddObject(type, vector2(x,y), zOrder, nameValuePairs, group);
	}

	// TODO: The onInit methods should be called after all game objects  
	//       have been added, and all properties have been set

	return 0;
}



