#include <assert.h>
#include <set>
#include <map>

#include "pathmgr.h"
#include "pathnode.h"
#include "scenemgr.h"
#include "application.h"
#include "acgfx_dynrender.h"
#include "pathedge.h"

// TODO: Evaluate if Detour can be used 
//       http://code.google.com/p/recastnavigation/

using namespace acUtility;
using namespace std;
using namespace acGraphics;

CPathMgr::CPathMgr()
{
	area   = 0;
	width  = 0;
	height = 0;
}

CPathMgr::~CPathMgr()
{
	if( area )
		delete[] area;
	for( unsigned int n = 0; n < nodes.size(); n++ )
		delete nodes[n];
	nodes.resize(0);
}

void CPathMgr::SetArea(unsigned int width, unsigned int height)
{
	if( area )
		delete[] area;
	for( unsigned int n = 0; n < nodes.size(); n++ )
		delete nodes[n];
	nodes.resize(0);

	area = new char[width*height];
	this->width  = width;
	this->height = height;
	memset(area, 0, width*height);
}

void CPathMgr::SetWalkable(unsigned int x, unsigned int y, bool walkable)
{
	if( x < width && y < height )
		area[x + y*width] = walkable ? 1 : 0;
}

vector2 CPathMgr::GetNearestPathNode(const vector2 &from, const vector2 &to)
{
	// Find the nodes that is in direct line of sight, those will be the starting paths to search

	return GetNearestUsingNavMesh(from, to);

//	return GetNearestUsingGrid(from, to);
}

// This function uses the navigation mesh to find a path
vector2 CPathMgr::GetNearestUsingNavMesh(const vector2 &from, const vector2 &to)
{
	// Go through each of the nodes until we find the one that encloses the from and to points
	// TODO: Speed up the search with using a hierarchical structure
	CPathNode *startNode = 0;
	CPathNode *endNode = 0;
	for( unsigned int n = 0; n < nodes.size(); n++ )
	{
		if( !startNode && nodes[n]->IsPointInNode(from) )
			startNode = nodes[n];
		if( !endNode && nodes[n]->IsPointInNode(to) )
			endNode = nodes[n];

		if( startNode && endNode )
			break;
	}

	// If the points of interest are not in the navigation mesh then we can't find a path
	if( !startNode || !endNode )
		return from;

	// If both points are in the same area, we can return immediately
	if( startNode == endNode )
		return startNode->ShiftPathToAvoidWall(from, to);

	struct SearchNode {
		SearchNode() { edge = 0; from = 0; realCost = 0; };
		SearchNode(const vector2 &_p, float _cost, float _rcost, CPathEdge *_edge, SearchNode *_from) { p = _p; estimatedCost = _cost; edge = _edge; from = _from; realCost = _rcost; }
		bool operator<(const SearchNode &o) const { if( p.x < o.p.x ) return true; if( p.x > o.p.x ) return false; if( p.y < o.p.y ) return true; return false; }
		vector2 p;
		float estimatedCost;
		float realCost;
		CPathEdge *edge;
		SearchNode *from;
	};
	// TODO: These should be members of the class to avoid having memory allocated with each call
	//       The class should also hold a pool of searchnodes for reuse 
	vector<SearchNode *> open;  // set of nodes to examine
	set<SearchNode*> visited;   // nodes already examined

	// Add each of the openings in the current node to the open search nodes
	CPathEdge *e = startNode->first;
	while( e )
	{
		if( e->neighbour )
		{
			vector2 mid = (e->p2 + e->p1)/2;
			float realCost = (mid - from).length();
			SearchNode *node = new SearchNode(mid, Heuristic(mid, to) + realCost, realCost, e, 0);
			open.push_back(node);
		}

		e = e->next;
		if( e == startNode->first )
			break;
	}
	
	while( open.size() > 0 )
	{
		// Get the least costly node from the open list
		// TODO: Should have a sorted set to make this access quick
		int best = 0;
		float bestCost = 1000;
		for( unsigned int n = 0; n < open.size(); n++ )
			if( bestCost > open[n]->estimatedCost )
			{
				bestCost = open[n]->estimatedCost;
				best = n;
			}
		SearchNode *curr = open[best];

		// Did we reach the goal?
		if( curr->edge->neighbour->owner == endNode )
		{
			// Smooth the path over the first few edges
			// to make it as straight as possible
			CPathEdge *edge[5] = {0};
			while( curr )
			{
				edge[4] = edge[3];
				edge[3] = edge[2];
				edge[2] = edge[1];
				edge[1] = curr->edge;
				// edge[0] will always be 0 to represent our starting point
				curr = curr->from;
			}

			vector2 p[5];
			p[4] = edge[4] ? (edge[4]->p1 + edge[4]->p2)/2 : to;
			p[3] = edge[3] ? (edge[3]->p1 + edge[3]->p2)/2 : to;
			p[2] = edge[2] ? (edge[2]->p1 + edge[2]->p2)/2 : to;
			p[1] = edge[1] ? (edge[1]->p1 + edge[1]->p2)/2 : to;
			p[0] = from;
			for( unsigned int i = 0; i < 3; i++ )
			{
				for( unsigned int n = 3; n > 0; n-- )
				{
					if( edge[n] == 0 ) continue;

					// Intersect line p[n+1] - p[n-1] on the edge[n]
					p[n] = intersectLines(p[n+1], p[n-1], edge[n]->p2, edge[n]->p1);

					// Shrink the edge with the radius of the moving object
					// TODO: Don't shrink if the opening is not sided be walls
					vector2 unit = edge[n]->p2 - edge[n]->p1;
					unit.normalize();
				
					// Project onto the edge to avoid moving too close to edges
					p[n] = projectPointOntoLineSegment(p[n], edge[n]->p2 - 0.45f*unit, edge[n]->p1 + 0.45f*unit);
				}
			}

			// Return the nearest node on the path
			// Set the waypoint slightly beyond the edge to avoid 
			// stopping immediately on top of the edge for the next
			// search, which may cause an infinite loop
			vector2 path = p[1] - 0.001f*edge[1]->GetNormal();

			// Make sure the path doesn't go too close to the wall
			path = startNode->ShiftPathToAvoidWall(from, path);

			// Clean up memory. All nodes in visited and open should be deleted
			for( set<SearchNode*>::iterator it = visited.begin(); it != visited.end(); it++ )
				delete *it;
			for( unsigned int n = 0; n < open.size(); n++ )
				delete open[n];

			return path;
		}

		// Remove the node from the open set
		open.erase(open.begin()+best);

		// Add it to the visited set
		visited.insert(curr);

		// Check each of the neighbours in the new node
		CPathNode *currNode = curr->edge->neighbour->owner;
		CPathEdge *e = currNode->first;
		while( e )
		{
			if( e->neighbour )
			{
				// Check if this edge hasn't been visited already
				// TODO: visited should be a set<CPathEdge*>
				bool isVisited = false;
				for( set<SearchNode*>::iterator it = visited.begin(); it != visited.end(); it++ )
				{
					if( (*it)->edge == e )
					{
						isVisited = true;
						break;
					}
				}

				if( !isVisited )
				{
					// The real cost to the neighbour is the cost from the current + to the neighbour
					vector2 mid = (e->p2 + e->p1)/2;
					float cost = curr->realCost + (mid - (curr->edge->p2 + curr->edge->p1)/2).length();

					// Look for the neighbour in the open set
					bool found = false;
					for( unsigned int o = 0; o < open.size(); o++ )
						if( open[o]->edge == e )
						{
							if( cost < open[o]->realCost )
							{
								// Update the estimated cost of the node as we've found a better path than previously found
								open[o]->estimatedCost = cost + Heuristic(mid, to);

								// Update the path used to get to this node
								open[o]->realCost = cost;
								open[o]->from = curr;
							}
							found = true;
							break;
						}

					// Add the neighbour to the open set with the estimated cost
					if( !found )
					{
						SearchNode *node = new SearchNode(mid, Heuristic(mid, to) + cost, cost, e, curr);
						open.push_back(node);
					}
				}
			}
			
			e = e->next;
			if( e == currNode->first )
				break;
		}
	}

	// Clean up memory. All nodes in visited should be deleted
	for( set<SearchNode*>::iterator it = visited.begin(); it != visited.end(); it++ )
		delete *it;

	// If no path can be found then return from vector
	return from;
}

// This function uses the grid directly to find a path
vector2 CPathMgr::GetNearestUsingGrid(const vector2 &from, const vector2 &to)
{
	// TODO: The path should be cached so that any new query for two coordinates on the path can reuse it

	// If outside the area, then just return the from vector
	if( to.x < 0 || to.x >= width ) return from;
	if( to.y < 0 || to.y >= height ) return from;

	// If the to vector is not walkable, then just return the from vector
	int toX = int(to.x);
	int toY = int(to.y);
	if( area[toX + toY*width] == 0 ) return from;
	
	// Run the A* algorithm to find the shortest path
	int fromX = int(from.x);
	int fromY = int(from.y);

	// TODO: These should be members of the path manager, to reduce the dynamic memory allocations
	struct node { 
		node() { x = 0; y = 0; cost = 0; }
		node(int x, int y, float cost) : x(x), y(y), cost(cost) {} 
		int x, y; 
		float cost; 
	};
	vector<node> open;                     // set of nodes to examine
	vector<bool> visited(width * height);  // nodes already examined
	vector<node> cameFrom(width * height); // backtracking to get the best path

	// Add the from node to the open set with the estimated cost to the target
	open.push_back(node(fromX, fromY, Heuristic(fromX, fromY, toX, toY)));

	// Add the from node to the cameFrom map with the actual cost from the from, i.e. zero
	cameFrom[fromX + fromY*width] = node(fromX, fromY, 0);

	while( open.size() > 0 )
	{
		// Get the least costly node from the open list
		// TODO: Should have a sorted set to make this access quick
		int best = 0;
		float bestCost = 1000;
		for( unsigned int n = 0; n < open.size(); n++ )
			if( bestCost > open[n].cost )
			{
				bestCost = open[n].cost;
				best = n;
			}
		node curr = open[best];

		// Did we reach the goal?
		if( curr.x == toX && curr.y == toY )
		{
			// Return the nearest node on the path
			for(;;)
			{
				node from = cameFrom[curr.x + curr.y*width];
				if( fromX == from.x && fromY == from.y )
				{
					if( curr.x == toX && curr.y == toY )
						return to;
					else
					{
						// TODO: Should remove redundant nodes from the path when the 
						//       next node is can be reached with a straight line
						return vector2(curr.x+0.5f, curr.y+0.5f);
					}
				}
				curr = from;
			}
		}

		// Remove the node from the open set
		open.erase(open.begin()+best);

		// Mark the node as visited so it isn't included again
		visited[curr.x + curr.y*width] = true;

		// Check each of the neighbours
		// TODO: Add diagonal neighbours too. But need to check if the path is free on both sides. 
		//       This won't be needed when changing to navigation mesh, as it will be non-uniform anyway
		int neighbours[][2] = {{1,0}, {0,1}, {0,-1}, {-1,0}};
		for( int n = 0; n < 4; n++ )
		{
			unsigned int nextX = curr.x + neighbours[n][0];
			unsigned int nextY = curr.y + neighbours[n][1];
			if( nextX >= width || nextY >= height ) continue;
			if( visited[nextX + nextY*width] || area[nextX + nextY*width] == 0 )
				continue;

			// The real cost to the neighbour is the cost from the current + to the neighbour
			float cost = cameFrom[curr.x + curr.y*width].cost + CostToNeighbour(curr.x, curr.y, nextX, nextY);

			// Look for the neighbour in the open set
			bool found = false;
			for( unsigned int o = 0; o < open.size(); o++ )
				if( open[o].x == nextX && open[o].y == nextY )
				{
					if( cost < cameFrom[nextX + nextY*width].cost )
					{
						// Update the estimated cost of the node as we've found a better path than previously found
						open[o].cost = cost + Heuristic(nextX, nextY, toX, toY);
					}
					found = true;
					break;
				}

			// Add the neighbour to the open set with the estimated cost
			if( !found )
				open.push_back(node(nextX, nextY, cost + Heuristic(nextX, nextY, toX, toY)));
		
			cameFrom[nextX + nextY*width] = node(curr.x, curr.y, cost);
		}
	}

	// If no path can be found then return from vector
	return from;
}

float CPathMgr::Heuristic(const vector2 &from, const vector2 &to)
{
	// Euclidian distance
	return (to - from).length();
}

float CPathMgr::Heuristic(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	// Euclidian distance
	return sqrtf(float((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1)));
}

float CPathMgr::CostToNeighbour(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	// Euclidian distance
	return sqrtf(float((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1)));
}

void CPathMgr::MakeRoom(const unsigned int x, const unsigned int y, bool *visited)
{
	// Initial size of the room
	unsigned int w = 1, h = 1;

	// TODO: Should avoid expanding long thin rectangles into open areas, 
	//       as then they cut the open area in multiple pieces

	// Try to expand the size as far as possible
	bool right = true, down = true;
	while( right || down )
	{
		// Check if it is possible to expand right
		if( right )
		{
			for( unsigned int y1 = y; y1 < y+h; y1++ )
			{
				if( !IsWalkable(x+w, y1) || visited[x+w + width*y1] )
				{
					right = false;
					break;
				}
			}
		}

		if( right )
			w++;

		// Check if it is possible to expand down
		if( down )
		{
			for( unsigned int x1 = x; x1 < x+w; x1++ )
			{
				if( !IsWalkable(x1, y+h) || visited[x1 + width*(y+h)] )
				{
					down = false;
					break;
				}
			}
		}

		if( down )
			h++;
	}

	// Fill the area as visited
	for( unsigned int y1 = y; y1 < y+h; y1++ )
		for( unsigned int x1 = x; x1 < x+w; x1++ )
			visited[x1 + y1*width] = true;

	// Create the path node
	CPathNode *node = new CPathNode(vector2(x+0.5f*w, y+0.5f*h));
	nodes.push_back(node);

	// Create the wall and door path edges
	CPathEdge *last = 0;

	// Wall to the north
	unsigned int s = x;
	bool open = IsWalkable(x, y-1);
	for( unsigned int x1 = x; x1 < x+w; x1++ )
	{
		if( IsWalkable(x1, y-1) != open )
		{
			CPathEdge *edge = new CPathEdge(vector2(float(s),float(y)), vector2(float(x1), float(y)), node);
			if( last == 0 )
			{
				last = edge;
				node->first = edge;
			}
			else
			{
				last->SetNext(edge);
				last = edge;
			}
			edge->wall = !open;
			open = !open;
			s = x1;
		}
	}

	CPathEdge *edge = new CPathEdge(vector2(float(s),float(y)), vector2(float(x+w), float(y)), node);
	if( last == 0 )
	{
		last = edge;
		node->first = edge;
	}
	else
	{
		last->SetNext(edge);
		last = edge;
	}
	edge->wall = !open;

	// Wall to the east
	s = y;
	open = IsWalkable(x+w, s);
	for( unsigned int y1 = s; y1 < y+h; y1++ )
	{
		if( IsWalkable(x+w, y1) != open )
		{
			CPathEdge *edge = new CPathEdge(vector2(float(x+w),float(s)), vector2(float(x+w), float(y1)), node);
			last->SetNext(edge);
			last = edge;
			edge->wall = !open;
			open = !open;
			s = y1;
		}
	}

	edge = new CPathEdge(vector2(float(x+w),float(s)), vector2(float(x+w), float(y+h)), node);
	last->SetNext(edge);
	last = edge;
	edge->wall = !open;

	// Wall to the south
	s = x+w-1;
	open = IsWalkable(s, y+h);
	for( unsigned int x1 = s; x1 >= x; x1-- )
	{
		if( IsWalkable(x1, y+h) != open )
		{
			CPathEdge *edge = new CPathEdge(vector2(float(s+1),float(y+h)), vector2(float(x1+1), float(y+h)), node);
			last->SetNext(edge);
			last = edge;
			edge->wall = !open;
			open = !open;
			s = x1;
		}
	}

	edge = new CPathEdge(vector2(float(s+1),float(y+h)), vector2(float(x), float(y+h)), node);
	last->SetNext(edge);
	last = edge;
	edge->wall = !open;

	// Wall to the west
	s = y+h-1;
	open = IsWalkable(x-1, s);
	for( unsigned int y1 = s; y1 >= y; y1-- )
	{
		if( IsWalkable(x-1, y1) != open )
		{
			CPathEdge *edge = new CPathEdge(vector2(float(x),float(s+1)), vector2(float(x), float(y1+1)), node);
			last->SetNext(edge);
			last = edge;
			edge->wall = !open;
			open = !open;
			s = y1;
		}
	}

	edge = new CPathEdge(vector2(float(x),float(s+1)), vector2(float(x), float(y)), node);
	last->SetNext(edge);
	last = edge;
	edge->wall = !open;

	// Connect the last edge with the first
	last->SetNext(node->first);
}

void CPathMgr::Optimize()
{
	// Process the grid to build a navigation mesh

	// Find rectangular areas as large as possible and form rooms out of them
	bool *visited = new bool[width*height];
	memset(visited, 0, sizeof(bool)*width*height);
	for( unsigned int y = 0; y < height; y++ )
	{
		for( unsigned int x = 0; x < width; x++ )
		{
			if( IsWalkable(x,y) && !visited[x + y*width] )
				MakeRoom(x,y, visited);
		}
	}
	delete[] visited;

	// For each of the nodes connect the open edges to the neighbouring nodes
	// It will be necessary to split open edges in case more than one neighbour is found
	vector<CPathNode*> unconnected = nodes;
	while( unconnected.size() )
	{
		for( unsigned int n1 = 0; n1 < unconnected.size(); n1++ )
		{
			CPathEdge *e1 = unconnected[n1]->first;
			bool hasUnconnected = false;
			for(;;)
			{
				if( !e1->wall && e1->neighbour == 0 )
				{
					// Find a neighbour to connect 
					for( unsigned int n2 = 0; n2 < unconnected.size(); n2++ )
					{
						CPathEdge *e2 = unconnected[n2]->first;
						for(;;)
						{
							if( !e2->wall && e2->neighbour == 0 && e2->p1 == e1->p2 )
							{
								if( e2->p2 == e1->p1 )
								{
									// Connect the edges
									e1->SetNeighbour(e2);
									break;
								}
								else
								{
									// Check if it is necessary to split the edge
									vector2 dir = e1->p2 - e1->p1;
									dir.normalize();
									vector2 nrm = vector2(dir.y, -dir.x);

									if( fabs(dot(nrm, e2->p1 - e2->p2)) < 0.0001f &&
										dot(dir, e2->p2 - e2->p1) < 0 )
									{
										// Which edge needs to be split?
										if( dot(dir, e1->p2 - e1->p1) > dot(dir, e2->p1 - e2->p2) )
										{
											CPathEdge *ne = e1->Split(e2->p2);
											ne->SetNeighbour(e2);
										}
										else
										{
											CPathEdge *ne = e2->Split(e1->p1);
											e1->SetNeighbour(e2);
											break;
										}
									}
								}
							}

							e2 = e2->next;
							if( e2 == unconnected[n2]->first )
								break;
						}

						if( e1->neighbour )
							break;
					}

					if( e1->neighbour == 0 )
						hasUnconnected = true;
				}

				e1 = e1->next;
				if( e1 == unconnected[n1]->first )
					break;
			}

			if( !hasUnconnected )
			{
				// This node is fully connected and can be removed
				unconnected.erase(unconnected.begin()+n1);
				n1--;
			}
		}
	}
}

bool CPathMgr::IsWalkable(unsigned int x, unsigned int y)
{
	if( x >= width ) return false;
	if( y >= height ) return false;
	return area[x + y*width] == 1;
}

void CPathMgr::DebugDraw()
{
	app->sceneMgr->PrepareRender();

	matrix4x4 mtx;
	mtx.setIdentity();
	app->gfx->SetMatrix(MTX_WORLD, mtx);

	app->gfx->SetTexture(0, 0);
	app->gfx->CommitEffectChanges();

	// Draw the nodes
	for( unsigned n = 0; n < nodes.size(); n++ )
	{
		app->render->Begin(RENDER_LINE_LIST);
		app->render->VtxColor(0x7F777700);

		app->render->VtxPos(nodes[n]->p.x - 0.4f, nodes[n]->p.y       , 0);
		app->render->VtxPos(nodes[n]->p.x       , nodes[n]->p.y - 0.4f, 0);

		app->render->VtxPos(nodes[n]->p.x       , nodes[n]->p.y - 0.4f, 0);
		app->render->VtxPos(nodes[n]->p.x + 0.4f, nodes[n]->p.y       , 0);
		
		app->render->VtxPos(nodes[n]->p.x + 0.4f, nodes[n]->p.y       , 0);
		app->render->VtxPos(nodes[n]->p.x       , nodes[n]->p.y + 0.4f, 0);
		
		app->render->VtxPos(nodes[n]->p.x       , nodes[n]->p.y + 0.4f, 0);
		app->render->VtxPos(nodes[n]->p.x - 0.4f, nodes[n]->p.y       , 0);

		// Draw edges
		app->render->VtxColor(0xFF000000);
		CPathEdge *edge = nodes[n]->first;
		while( edge )
		{
			if( edge->wall )
				app->render->VtxColor(0xFF000000);
			else
				app->render->VtxColor(0x7F770000);
			app->render->VtxPos(edge->p1.x, edge->p1.y, 0);
			app->render->VtxPos(edge->p2.x, edge->p2.y, 0);

			if( !edge->wall && edge->neighbour )
			{
				app->render->VtxColor(0x44777700);
				app->render->VtxPos(nodes[n]->p.x, nodes[n]->p.y, 0);
				app->render->VtxPos(((edge->p2+edge->p1)/2).x, ((edge->p2+edge->p1)/2).y, 0);
			}

			edge = edge->next;
			if( edge == nodes[n]->first )
				break;
		}

		// Draw the connections
/*		app->render->VtxColor(0x7FFF0000);
		for( unsigned c = 0; c < nodes[n]->connections.size(); c++ )
		{
			vector2 nrm(nodes[n]->connections[c]->p.x - nodes[n]->p.x, 
				        nodes[n]->connections[c]->p.y - nodes[n]->p.y);
			nrm.normalize();

			nrm = vector2(nrm.y, -nrm.x);
			nrm *= 0.2f;
			
			app->render->VtxPos(nodes[n]->p.x - nrm.x, nodes[n]->p.y - nrm.y, 0);
			app->render->VtxPos(nodes[n]->connections[c]->p.x - nrm.x, nodes[n]->connections[c]->p.y - nrm.y, 0);
		}
*/
		app->render->End();
	}
}