#include "physicsmgr.h"
#include "physicsobject.h"
#include "acgfx_dynrender.h"
#include "acutil_math.h"
#include "application.h"
#include "gamemgr.h"
#include <box2d.h>
#include "scenemgr.h"

using namespace acUtility;
using namespace acGraphics;
using namespace std;

class CContactListener : public b2ContactListener
{
public:
	CContactListener(CPhysicsMgr *mgr) : b2ContactListener() {this->mgr = mgr;}

	bool IsOneWayPassthrough(b2Fixture *f)
	{
		b2Shape *shape = f->GetShape();
		if( shape->m_type != b2Shape::e_edge ) return false;

		return true;
	}

	void CheckSensor(b2Fixture *fsensor, b2Fixture *fobject, bool doBegin)
	{
		b2Body *sensor = fsensor->GetBody();
		b2Body *object = fobject->GetBody();

		// Before we register the event we must check if the body 
		// is in contact with the sensor in some other fixture
		b2ContactEdge *contacts = object->GetContactList();
		bool inContact = false;
		while( contacts )
		{
			if( contacts->other == sensor &&
				contacts->contact->IsTouching() &&
				contacts->contact->GetFixtureA() != fobject &&
				contacts->contact->GetFixtureB() != fobject )
			{
				inContact = true;
				break;
			}

			contacts = contacts->next;
		}

		if( !inContact )
		{
			CPhysicsMgr::SSensorProximity prox = 
			{
				reinterpret_cast<CPhysicsObject*>(sensor->GetUserData()),
				reinterpret_cast<CPhysicsObject*>(object->GetUserData()),
				doBegin
			};
			mgr->sensorProximities.push_back(prox);
		}
	}

	void BeginContact(b2Contact* contact) 
	{
		// Is this a sensor?
		if( contact->GetFixtureA()->IsSensor() && !contact->GetFixtureB()->IsSensor() )
			CheckSensor(contact->GetFixtureA(), contact->GetFixtureB(), true);
		else if( contact->GetFixtureB()->IsSensor() && !contact->GetFixtureA()->IsSensor() )
			CheckSensor(contact->GetFixtureB(), contact->GetFixtureA(), true);

		// Count the number of contacts for the body (not fixture)
		if( IsOneWayPassthrough(contact->GetFixtureA()) )
		{
			b2Body *b = contact->GetFixtureB()->GetBody();
			for( unsigned int n = 0; n < mgr->ignoreContacts.size(); n++ )
			{
				if( mgr->ignoreContacts[n] == b )
				{
					++mgr->ignoreCount[n];
					break;
				}
			}
		}
	}
	void EndContact(b2Contact* contact) 
	{
		// Is this a sensor?
		if( contact->GetFixtureA()->IsSensor() && !contact->GetFixtureB()->IsSensor() )
			CheckSensor(contact->GetFixtureA(), contact->GetFixtureB(), false);
		else if( contact->GetFixtureB()->IsSensor() && !contact->GetFixtureA()->IsSensor() )
			CheckSensor(contact->GetFixtureB(), contact->GetFixtureA(), false);

		// Remove the contact from the ignore list
		if( IsOneWayPassthrough(contact->GetFixtureA()) )
		{
			b2Body *b = contact->GetFixtureB()->GetBody();
			for( unsigned int n = 0; n < mgr->ignoreContacts.size(); n++ )
			{
				if( mgr->ignoreContacts[n] == b )
				{
					if( --mgr->ignoreCount[n] == 0 )
					{
						mgr->ignoreContacts[n] = *mgr->ignoreContacts.rbegin();
						mgr->ignoreContacts.pop_back();
						mgr->ignoreCount[n] = *mgr->ignoreCount.rbegin();
						mgr->ignoreCount.pop_back();
					}
					break;
				}
			}
		}
	}

	void PreSolve(b2Contact* contact, const b2Manifold* oldManifold)
	{ 
		b2WorldManifold worldManifold;
		contact->GetWorldManifold(&worldManifold);

		// TODO: Should have proper user data, or could use the filter 
		//       category to identify the one way passthrough.
		// TODO: The ignorelist should be stored in the physics object attached to the platform

		// Is this a one-way passthrough platform?
		// Only edge shapes with 2 vertices can be one-way passthrough
		b2Shape *shape = contact->GetFixtureA()->GetShape();
		if( shape->m_type == b2Shape::e_edge )
		{
			// Check if the contact should be ignored
			b2Body *b = contact->GetFixtureB()->GetBody();
			for( unsigned int n = 0; n < mgr->ignoreContacts.size(); n++ )
				if( mgr->ignoreContacts[n] == b )
				{
					contact->SetEnabled(false);
					return;
				}

			// Determine the normal of the platform
			b2EdgeShape *platformShape = dynamic_cast<b2EdgeShape*>(shape);
			b2Vec2 norm = platformShape->m_vertex2 - platformShape->m_vertex1;
			norm = b2Vec2(-norm.y, norm.x);
			norm.Normalize();

			// TODO: bug: There is a problem here. If the first contact to be verified happens
			//            to be slightly below then the entire object passes through. Due to
			//            numerical errors this can allow an object to slip through. 

			// TODO: If any part of the body (all fixtures) is below the platform when
			//        the first contact is established, the collision should be ignored

			// If the movement is in the same general direction as the normal
			// of the platform then the object should pass through
			b2Vec2 vA = contact->GetFixtureA()->GetBody()->GetLinearVelocityFromWorldPoint(worldManifold.points[0]);
			b2Vec2 vB = contact->GetFixtureB()->GetBody()->GetLinearVelocityFromWorldPoint(worldManifold.points[0]);
//			move.Normalize();
			if( b2Dot(norm, vB-vA) > 0.8f )
			{
				contact->SetEnabled(false);
				mgr->ignoreContacts.push_back(b);
				mgr->ignoreCount.push_back(1);
				return;
			}
		}

		// Is it a new collision?
		b2PointState state1[2], state2[2];
		b2GetPointStates(state1, state2, oldManifold, contact->GetManifold());
		if( state2[0] == b2_addState )
		{
			b2Vec2 vA = contact->GetFixtureA()->GetBody()->GetLinearVelocityFromWorldPoint(worldManifold.points[0]);
			b2Vec2 vB = contact->GetFixtureB()->GetBody()->GetLinearVelocityFromWorldPoint(worldManifold.points[0]);
			float32 approachVelocity = -b2Dot(vB-vA, worldManifold.normal);
			
			if( approachVelocity > 0.1f )
			{
				// Store the collision
				CPhysicsMgr::SCollision coll = 
				{
					reinterpret_cast<CPhysicsObject*>(contact->GetFixtureA()->GetBody()->GetUserData()),
					reinterpret_cast<CPhysicsObject*>(contact->GetFixtureB()->GetBody()->GetUserData()),
					*reinterpret_cast<vector2*>(&worldManifold.points[0]),
					-*reinterpret_cast<vector2*>(&worldManifold.normal),
					approachVelocity
				};
			
				mgr->collisions.push_back(coll);
			}
		}
	}

	void PostSolve(b2Contact* contact, const b2ContactImpulse* impulse)
	{
		// We don't care about the impulse at this moment so we skip this
	}

	CPhysicsMgr *mgr;
};

CPhysicsMgr::CPhysicsMgr()
{
	world = 0;
	gravity = vector2(0,0);
}

CPhysicsMgr::~CPhysicsMgr()
{
	Uninit();
}

int CPhysicsMgr::Init()
{
	world = new b2World(b2Vec2(gravity.x, gravity.y));

	nextFrame = 0;

	return 0;
}

void CPhysicsMgr::SetGlobalGravity(const acUtility::vector2 &vec)
{
	gravity = vec;
	world->SetGravity(b2Vec2(vec.x, vec.y));
}

void CPhysicsMgr::Uninit()
{
	if( world )
	{
		delete world;
		world = 0;
	}
}

void CPhysicsMgr::ClearWorld()
{
	Uninit();
	Init();
}

void CPhysicsMgr::Frame()
{
	collisions.clear();
	sensorProximities.clear();

	CContactListener listener(this);
	world->SetContactListener(&listener);

	// Apply the gravity force
	for( unsigned int n = 0; n < objects.size(); n++ )
	{
		if( objects[n]->m_useCustomGravity )
		{
			b2Body *body = objects[n]->m_body;

			// Cancel the world's gravity
			body->ApplyForce(body->GetMass() * -world->GetGravity(), body->GetWorldCenter());

			// Apply the custom gravity
			body->ApplyForce(*reinterpret_cast<b2Vec2*>(&(body->GetMass() * objects[n]->m_gravity)), body->GetWorldCenter());
		}
	}

	// TODO: The timestep should be fixed (requires the rendering to be interpolated)
/*	int steps = int(100*(nextFrame + app->gameMgr->GetFrameDeltaTime()));
	nextFrame = app->gameMgr->GetFrameDeltaTime() - steps*0.01f;
	while( steps-- )
		world->Step(0.01f, 10, 10);
*/

	// TODO: Doing 10 iterations allow the objects on the platform to stay on it
	int iterations = 1;
	for( int n = 0; n < iterations; n++ )
		world->Step(app->gameMgr->GetFrameDeltaTime()/float(iterations), 10, 10);

	world->SetContactListener(0);
}

bool CPhysicsMgr::CanObjectMoveTo(const vector2 &pos, const CPhysicsObject *obj)
{
	bool canMove = true;

	// TODO: Create a sensor fixture with the same size as the object, 
	//       insert it in the world and determine the objects it touches

	return canMove;
}

class CRayCastCallback : public b2RayCastCallback
{
public:
	CRayCastCallback(const CPhysicsObject *ignore, int query) : b2RayCastCallback()
	{
		ignoreObject = ignore;
		queryTypes   = query;
		timeOfHit    = 1;
		firstObject  = 0;
	}

	float32 ReportFixture(b2Fixture* fixture, const b2Vec2& point,
						  const b2Vec2& normal, float32 fraction)
	{
		// We ignore all sensors in this raycast
		if( fixture->IsSensor() )
			return -1;

		// If a specific object should be ignored, then we do that
		if( ignoreObject && fixture->GetBody()->GetUserData() == ignoreObject )
			return -1;

		// If queryTypes == 1 we only want solid objects
		if( queryTypes == 1 && fixture->GetFilterData().maskBits == 0 )
			return -1;

		firstObject = fixture;
		timeOfHit   = fraction;

		return fraction;
	}

	const CPhysicsObject *ignoreObject;
	int                   queryTypes;
	float                 timeOfHit;
	b2Fixture            *firstObject;
};

float CPhysicsMgr::RayCast(const CPhysicsObject *ignore, int query, const acUtility::vector2 &v1, const acUtility::vector2 &v2, CPhysicsObject **firstContact)
{
	// TODO: Should return normal of the object hit
	// TODO: Should be possible to give a flag for types of objects to ignore, for example sensors

	CRayCastCallback callback(ignore, query);
	world->RayCast(&callback, *reinterpret_cast<const b2Vec2*>(&v1), *reinterpret_cast<const b2Vec2*>(&v2));

	if( firstContact && callback.firstObject )
	{
		*firstContact = reinterpret_cast<CPhysicsObject*>(callback.firstObject->GetBody()->GetUserData());
	}

	return callback.timeOfHit;
}


CPhysicsObject *CPhysicsMgr::AddCollisionLine(const vector2 &v1, const vector2 &v2, bool isDynamic)
{
	// TODO: should have a single world body for all static data
	vector2 midPoint = 0.5f*(v1+v2);
	b2BodyDef def;
	def.type = isDynamic ? b2_kinematicBody : b2_staticBody;
	def.position.Set(midPoint.x, midPoint.y);
	b2Body *body = world->CreateBody(&def);
	b2EdgeShape shape;
	shape.Set(*reinterpret_cast<b2Vec2*>(&(v1-midPoint)), *reinterpret_cast<b2Vec2*>(&(v2-midPoint)));

	b2FixtureDef fixtureDef;
	fixtureDef.shape = &shape;
	fixtureDef.density = 0; // static and kinematic bodies don't have a weight
	fixtureDef.friction = 1; // TODO: Make configurable
	fixtureDef.restitution = 0; // TODO: Make configurable

	body->CreateFixture(&fixtureDef);

	if( isDynamic )
	{
		CPhysicsObject *obj = new CPhysicsObject();
		obj->m_body = body;
		body->SetUserData(obj);
		objects.push_back(obj);
		return obj;
	}

	return 0;
}

CPhysicsObject *CPhysicsMgr::AddCollisionBox(const vector2 &v1, const vector2 &v2, bool isDynamic)
{
	// TODO: should have a single world body for all static data
	vector2 midPoint = 0.5f*(v1+v2);
	b2BodyDef def;
	def.type = isDynamic ? b2_kinematicBody : b2_staticBody;
	def.position.Set(midPoint.x, midPoint.y);
	b2Body *body = world->CreateBody(&def);
	b2PolygonShape shape;
	shape.SetAsBox(0.5f*fabs(v2.x-v1.x), 0.5f*fabs(v2.y-v1.y));

	b2FixtureDef fixtureDef;
	fixtureDef.shape = &shape;
	fixtureDef.density = 0; // static and kinematic bodies don't have a weight
	fixtureDef.friction = 1; // TODO: Make configurable
	fixtureDef.restitution = 0; // TODO: Make configurable

	body->CreateFixture(&fixtureDef);

	if( isDynamic )
	{
		CPhysicsObject *obj = new CPhysicsObject();
		obj->m_body = body;
		body->SetUserData(obj);
		objects.push_back(obj);
		return obj;
	}

	return 0;
}

class CDebugDrawer : public b2Draw 
{
public:
	CDebugDrawer() : b2Draw() 
	{
		SetFlags(0xFFFFFFFF);

		app->sceneMgr->PrepareRender();

		matrix4x4 mtx;
		mtx.setIdentity();
		app->gfx->SetMatrix(MTX_WORLD, mtx);

		app->gfx->SetTexture(0, 0);
		app->gfx->CommitEffectChanges();
	}

	void DrawPolygon(const b2Vec2* vertices, int32 vertexCount, const b2Color& color)
	{
		app->render->Begin(RENDER_LINE_LIST);
		app->render->VtxColor(f2d(color, false));
		for( int n = 0; n < vertexCount; n++ )
		{
			vector2 pos1(vertices[n].x, vertices[n].y);
			app->render->VtxPos(pos1.x, pos1.y, 0);
			vector2 pos2(vertices[(n+1)%vertexCount].x, vertices[(n+1)%vertexCount].y);
			app->render->VtxPos(pos2.x, pos2.y, 0);
		}
		app->render->End();
	}

	void DrawSolidPolygon(const b2Vec2* vertices, int32 vertexCount, const b2Color& color)
	{
		app->render->Begin(RENDER_TRIANGLE_LIST);
		app->render->VtxColor(f2d(color, true));
		for( int n = 1; n < vertexCount; n++ )
		{
			vector2 pos0(vertices[0].x, vertices[0].y);
			app->render->VtxPos(pos0.x, pos0.y, 0);
			vector2 pos1(vertices[n].x, vertices[n].y);
			app->render->VtxPos(pos1.x, pos1.y, 0);
			vector2 pos2(vertices[(n+1)%vertexCount].x, vertices[(n+1)%vertexCount].y);
			app->render->VtxPos(pos2.x, pos2.y, 0);
		}
		app->render->End();
		DrawPolygon(vertices, vertexCount, color);
	}

	void DrawCircle(const b2Vec2& center, float32 radius, const b2Color& color)
	{
		app->render->Begin(RENDER_LINE_LIST);
		app->render->VtxColor(f2d(color, false));
		for( int n = 0; n < 16; n++ )
		{
			vector2 pos(center.x, center.y);
			app->render->VtxPos(pos.x + radius*cos(n*2*pi/16), pos.y + radius*sin(n*2*pi/16), 0);
			app->render->VtxPos(pos.x + radius*cos((n+1)*2*pi/16), pos.y + radius*sin((n+1)*2*pi/16), 0);
		}
		app->render->End();
	}
	
	void DrawSolidCircle(const b2Vec2& center, float32 radius, const b2Vec2& axis, const b2Color& color)
	{
		app->render->Begin(RENDER_TRIANGLE_LIST);
		app->render->VtxColor(f2d(color, true));
		for( int n = 0; n < 16; n++ )
		{
			vector2 pos(center.x, center.y);
			app->render->VtxPos(pos.x, pos.y, 0);
			app->render->VtxPos(pos.x + radius*cos(n*2*pi/16), pos.y + radius*sin(n*2*pi/16), 0);
			app->render->VtxPos(pos.x + radius*cos((n+1)*2*pi/16), pos.y + radius*sin((n+1)*2*pi/16), 0);
		}
		app->render->End();
		DrawCircle(center, radius, color);
	}
	
	void DrawSegment(const b2Vec2& p1, const b2Vec2& p2, const b2Color& color)
	{
		app->render->Begin(RENDER_LINE_LIST);
		app->render->VtxColor(f2d(color, false));

		vector2 pos1(p1.x, p1.y);
		app->render->VtxPos(pos1.x, pos1.y, 0);
		vector2 pos2(p2.x, p2.y);
		app->render->VtxPos(pos2.x, pos2.y, 0);

		app->render->End();
	}

	void DrawTransform(const b2Transform& xf)
	{
		app->render->Begin(RENDER_LINE_LIST);
		app->render->VtxColor(0x7FFFFF00);

		vector2 pos1(xf.p.x, xf.p.y);
		app->render->VtxPos(pos1.x, pos1.y, 0);
		vector2 pos2(xf.p.x + cosf(xf.q.GetAngle()), xf.p.y + sinf(xf.q.GetAngle()));
		app->render->VtxPos(pos2.x, pos2.y, 0);
		vector2 pos3(xf.p.x - sinf(xf.q.GetAngle()), xf.p.y + cosf(xf.q.GetAngle()));
		app->render->VtxPos(pos1.x, pos1.y, 0);
		app->render->VtxPos(pos3.x, pos3.y, 0);

		app->render->End();
	}

	UINT f2d(const b2Color &f, bool solid)
	{
		UINT d = solid ? 0x7F000000 : 0xFF000000;
		d |= (int(f.r*255)&0xFF)<<16;
		d |= (int(f.g*255)&0xFF)<<8;
		d |= (int(f.b*255)&0xFF)<<0;
		return d;
	}
};

void CPhysicsMgr::DebugDraw()
{
	CDebugDrawer draw;
	world->SetDebugDraw(&draw);
	world->DrawDebugData();
	world->SetDebugDraw(0);
}

CPhysicsObject *CPhysicsMgr::CreateObjectAsBullet(const vector2 &pos, float weight, EPhysicsProperty prop)
{
	CPhysicsObject *obj = new CPhysicsObject();
	objects.push_back(obj);
	obj->ClearCustomGravity();

	b2BodyDef def;
	def.userData = obj;
	def.type = b2_dynamicBody;
	def.position.Set(pos.x, pos.y);
	def.bullet = true;
	b2Body *body = world->CreateBody(&def);
	
	b2CircleShape shape;
	shape.m_radius = 0.01f;

	b2FixtureDef fixtureDef;
	fixtureDef.shape = &shape;
	fixtureDef.density = weight/(pi*0.01f*0.01f); // density is kg/m^2
	fixtureDef.friction = 0.75f;
	fixtureDef.restitution = 0; // TODO: Make configurable
	fixtureDef.filter.maskBits = prop == PP_PASSTHROUGH ? 0 : 0xFFFF;
	fixtureDef.isSensor = prop == PP_SENSOR ? true : false;
	body->CreateFixture(&fixtureDef);

	obj->m_body = body;

	return obj;
}

CPhysicsObject *CPhysicsMgr::CreateObjectAsCircle(const vector2 &pos, float collisionRadius, float weight, EPhysicsProperty prop)
{
	CPhysicsObject *obj = new CPhysicsObject();
	objects.push_back(obj);
	obj->ClearCustomGravity();

	b2BodyDef def;
	def.userData = obj;
	def.type = weight == 0 ? b2_kinematicBody : b2_dynamicBody;
	def.position.Set(pos.x, pos.y);
	def.fixedRotation = false; // TODO: This should be selectable 
	b2Body *body = world->CreateBody(&def);
	b2CircleShape shape;
	shape.m_radius = collisionRadius;
	b2FixtureDef fixtureDef;
	fixtureDef.shape = &shape;
	fixtureDef.density = weight/(pi*collisionRadius*collisionRadius); // density is kg/m^2
	fixtureDef.friction = 0.75f;
	fixtureDef.restitution = 0; // TODO: Make configurable
	fixtureDef.filter.maskBits = prop == PP_PASSTHROUGH ? 0 : 0xFFFF;
	fixtureDef.isSensor = prop == PP_SENSOR ? true : false;
	body->CreateFixture(&fixtureDef);

	obj->m_body = body;

	return obj;
}

CPhysicsObject *CPhysicsMgr::CreateObjectAsBox(const vector2 &pos, float halfWidth, float halfHeight, float weight, EPhysicsProperty prop)
{
	CPhysicsObject *obj = new CPhysicsObject();
	objects.push_back(obj);
	obj->ClearCustomGravity();

	b2BodyDef def;
	def.userData = obj;
	def.type = weight == 0 ? b2_kinematicBody : b2_dynamicBody;
	def.position.Set(pos.x, pos.y);
	def.fixedRotation = false; // TODO: This should be selectable 
	b2Body *body = world->CreateBody(&def);
	b2PolygonShape shape;
	shape.SetAsBox(halfWidth, halfHeight);
	b2FixtureDef fixtureDef;
	fixtureDef.shape = &shape;
	fixtureDef.density = weight/(2*halfWidth * 2*halfHeight); // density is kg/m^2
	fixtureDef.friction = 0.75f; // TODO: make configurable
	fixtureDef.restitution = 0; // TODO: Make configurable
	fixtureDef.filter.maskBits = prop == PP_PASSTHROUGH ? 0 : 0xFFFF;
	fixtureDef.isSensor = prop == PP_SENSOR ? true : false;
	body->CreateFixture(&fixtureDef);

	obj->m_body = body;

	return obj;
}

CPhysicsObject *CPhysicsMgr::CreateObjectAsCapsule(const vector2 &pos, float halfWidth, float halfHeight, float weight, EPhysicsProperty prop)
{
	CPhysicsObject *obj = new CPhysicsObject();
	objects.push_back(obj);
	obj->ClearCustomGravity();

	b2BodyDef def;
	def.userData = obj;
	def.type = weight == 0 ? b2_kinematicBody : b2_dynamicBody;
	def.position.Set(pos.x, pos.y);
	def.fixedRotation = true; // TODO: This should be selectable 
	b2Body *body = world->CreateBody(&def);
	obj->m_body = body;

	b2FixtureDef fixtureDef;
	fixtureDef.friction = 0.0f; // TODO: Friction should be configurable, but 0 is recommended for objects that are manually controlled, like the player (otherwise the object may stick to walls, if the velocity is perpendicular to the wall)
	fixtureDef.restitution = 0; // TODO: Make configurable
	fixtureDef.density = weight/((2*halfWidth*2*(halfHeight-halfWidth))+(pi*halfWidth*halfWidth));
	fixtureDef.filter.maskBits = prop == PP_PASSTHROUGH ? 0 : 0xFFFF;
	fixtureDef.isSensor = prop == PP_SENSOR ? true : false;

	if( halfWidth < halfHeight )
	{
		b2PolygonShape box;
		// Make the box slightly smaller than the circles to avoid the capsule 
		// getting caught on corners due to numeric rounding errors
		box.SetAsBox(halfWidth*0.95f, halfHeight-halfWidth);
		fixtureDef.shape = &box;

		body->CreateFixture(&fixtureDef);

		b2CircleShape circle;
		circle.m_radius = halfWidth;
		circle.m_p = b2Vec2(0,halfHeight-halfWidth);
		fixtureDef.shape = &circle;
		body->CreateFixture(&fixtureDef);

		circle.m_p = -circle.m_p;
		body->CreateFixture(&fixtureDef);
	}
	else
	{
		b2PolygonShape box;
		// Make the box slightly smaller than the circles to avoid the capsule 
		// getting caught on corners due to numeric rounding errors
		box.SetAsBox(halfWidth-halfHeight, halfHeight*0.95f);
		body->CreateFixture(&fixtureDef);

		b2CircleShape circle;
		circle.m_radius = halfHeight;
		circle.m_p = b2Vec2(halfWidth-halfHeight,0);
		fixtureDef.shape = &circle;
		body->CreateFixture(&fixtureDef);

		circle.m_p = -circle.m_p;
		body->CreateFixture(&fixtureDef);
	}

	return obj;
}

void CPhysicsMgr::DestroyObject(CPhysicsObject *obj)
{
	vector<CPhysicsObject*>::const_iterator i = objects.begin();
	while( i != objects.end() )
	{
		if( *i == obj )
		{
			objects.erase(i);
			break;
		}
		i++;
	}

	if( obj->m_body )
		world->DestroyBody(obj->m_body);

	delete obj;
}

int CPhysicsMgr::GetCollisionCount()
{
	return collisions.size();
}

void CPhysicsMgr::GetCollision(int i, CPhysicsObject **a, CPhysicsObject **b, vector2 *pos, vector2 *norm, float *impulse)
{
	if( i < 0 || i >= (int)collisions.size() )
	{
		if( a ) *a = 0;
		if( b ) *b = 0;
		if( pos ) *pos = vector2(0,0);
		if( norm ) *norm = vector2(0,0);
		if( impulse ) *impulse = 0;
	}
	else
	{
		if( a ) *a = collisions[i].a;
		if( b ) *b = collisions[i].b;
		if( pos ) *pos = collisions[i].pos;
		if( norm ) *norm = collisions[i].norm;
		if( impulse ) *impulse = collisions[i].impulse;
	}
}

int CPhysicsMgr::GetSensorProximityCount()
{
	return sensorProximities.size();
}

void CPhysicsMgr::GetSensorProximity(int i, CPhysicsObject **a, CPhysicsObject **b, bool *isNearing)
{
	if( i < 0 || i >= (int)sensorProximities.size() )
	{
		if( a ) *a = 0;
		if( b ) *b = 0;
		if( isNearing ) *isNearing = false;
	}
	else
	{
		if( a ) *a = sensorProximities[i].sensor;
		if( b ) *b = sensorProximities[i].object;
		if( isNearing ) *isNearing = sensorProximities[i].isNearing;
	}
}