#include "recorder.h"
#include "application.h"
#include "gamemgr.h"
#include "acutil_input.h"

#include <string>
#include <sstream>
using namespace std;

// TODO: Recorder enhancements
//       - It should be possible to in/decrease the playback speed with the arrows. This will simply control how many frames should be updated between each render
//       - By pressing space, the playback should be paused and stepped through frame by frame
//       - Pressing escape will leave the recording early and exit the game
//       - The recorder should also display the current frame on the screen as well as the speed for the playback
//       - It would even be possible to allow setting breakpoints in the play file

CRecorder::CRecorder()
{
	memset(keyState, 0, 256);
	memset(mbState, 0, 8);
	memset(actionState, 0, 256);
	frame = 0;
	playback = 0;
	cursorX = 0;
	cursorY = 0;
}

CRecorder::~CRecorder()
{
	if( playback )
		fclose(playback);
}

void CRecorder::RecordSetup()
{
	// Save the initial state such as seed for randomizer, etc
	FILE *f;
	fopen_s(&f, "play.rec", "w");

	fprintf(f, "seed=%u\n", app->gameMgr->GetRandomSeed());
	
	fclose(f);
}

void CRecorder::RecordFrame()
{
	frame++;

	// Determine what has changed since last frame. 
	stringstream changes;
	for( UINT n = 0; n < 256; n++ )
	{
		int state = app->input->GetKeyState(n);
		if( state != keyState[n] )
		{
			changes << "k" << n << "=" << state << ",";
			keyState[n] = state;
		}
	}

	for( UINT n = 0; n < 8; n++ )
	{
		int state = app->input->GetMouseButtonState(n);
		if( state != mbState[n] )
		{
			changes << "mb" << n << "=" << state << ",";
			mbState[n] = state;
		}
	}

	for( UINT n = 0; n < 256; n++ )
	{
		int state = app->input->GetActionState(n) ? (0x80 + (app->input->DidActionStart(n) ? 0 : 1)) : 0;
		if( state != actionState[n] )
		{
			changes << "a" << n << "=" << state << ",";
			actionState[n] = state;
		}
	}

	int dx = 0, dy = 0, dz = 0;
	app->input->GetMouseDelta(&dx, &dy, &dz);
	if( dx )
		changes << "dx=" << dx << ",";
	if( dy )
		changes << "dy=" << dy << ",";
	if( dz )
		changes << "dz=" << dz << ",";

	int x,y;
	app->input->GetCursorPos(&x, &y);
	if( cursorX != x )
	{
		changes << "cx=" << x << ",";
		cursorX = x;		
	}
	if( cursorY != y )
	{
		changes << "cy=" << y << ",";
		cursorY = y;
	}

	// Only save an entry if something changed
	if( changes.str().length() )
	{
		FILE *f;
		fopen_s(&f, "play.rec", "a");
		fprintf(f, "%d:%s\n", frame, changes.str().c_str());
		fclose(f);
	}
}

void CRecorder::PlaybackSetup()
{
	// When doing a playback we open the file at setup and keep it open
	fopen_s(&playback, "play.rec", "r");
	if( playback == 0 )
		return;

	char line[1000];
	fgets(line, 1000, playback);
	UINT seed = 0;
	sscanf_s(line, "seed=%u", &seed);

	app->gameMgr->SetRandomSeed(seed);
}

void CRecorder::PlaybackFrame()
{
	frame++;

	if( playback == 0 )
		return;

	int dx = 0, dy = 0, dz = 0;

	// Remember current position in case the current frame doesn't have any changes
	long pos = ftell(playback);

	char line[1000];
	fgets(line, 1000, playback);
	UINT f = 0;
	sscanf_s(line, "%u:", &f);

	if( f != frame )
	{
		fseek(playback, pos, SEEK_SET);
	}
	else
	{
		// Scan the recording to update the internal state
		char *token = strchr(line, ':');
		if( token )
		{
			token++;
			for(;;)
			{
				string name, value;
				char *e = strchr(token, '=');
				if( !e ) break;
				name.assign(token, e);
				token = e + 1;

				e = strchr(token, ',');
				if( !e ) break;
				value.assign(token, e);
				token = e + 1;

				// Update the state
				if( name == "dx" )
					dx = atoi(value.c_str());
				else if( name == "dy" )
					dy = atoi(value.c_str());
				else if( name == "dz" )
					dz = atoi(value.c_str());
				else if( name.substr(0, 2) == "mb" )
				{
					int button = name[2] - '0';
					int state = atoi(value.c_str());

					mbState[button] = state;
				}
				else if( name[0] == 'k' )
				{
					int key = atoi(&name[1]);
					int state = atoi(value.c_str());

					keyState[key] = state;
				}
				else if( name[0] == 'a' )
				{
					int action = atoi(&name[1]);
					int state = atoi(value.c_str());

					actionState[action] = state;
				}
				else if( name == "cx" )
					cursorX = atoi(value.c_str());
				else if( name == "cy" )
					cursorY = atoi(value.c_str());
			}
		}
	}

	// Push the state onto the input manager
	acUtility::CInput *input = app->input;
	input->SetMouseDelta(dx, dy, dz);
	for( UINT n = 0; n < 8; n++ )
		input->SetMouseButtonState(n, mbState[n]);
	for( UINT n = 0; n < 256; n++ )
		input->SetKeyState(n, keyState[n]);
	for( UINT n = 0; n < 256; n++ )
		input->SetActionState(n, actionState[n]);
	input->SetCursorPos(cursorX, cursorY);
}