#include <assert.h>
#include "scenemgr.h"
#include "acgfx_graphics.h"
#include "sprite.h"
#include "material.h"
#include "acutil_log.h"
#include "application.h"
#include "acutil_filemgr.h"
#include "acgfx_dynrender.h"
#include "acutil_config.h"
#include <algorithm>


using namespace acGraphics;
using namespace acUtility;
using namespace std;


CSceneMgr::CSceneMgr()
{
	initialized = false;
	cameraPos = vector2(0,0);
	cameraZoom = 1;
	fixedViewport = false;
}

CSceneMgr::~CSceneMgr()
{
	Uninit();
}

int CSceneMgr::Init(CGraphics *gfx)
{
	this->gfx = gfx;

	// Load the effects that are needed for rendering the scene
	fxScene = gfx->LoadEffect(app->fileMgr->GetPathForFile("scene.fx").c_str());

	gfx->RegisterDeviceUser(this);

	ReinitializeDeviceObjects(gfx);

	if( fxScene < 0 )
		return -1;

	return 0;
}

void CSceneMgr::ReinitializeDeviceObjects(CGraphics *gfx)
{
	initialized = true;
}

void CSceneMgr::ReleaseDeviceObjects()
{
	initialized = 0;
}

void CSceneMgr::ClearScene()
{
	for( unsigned int n = 0; n < sprites.size(); n++ )
		if( sprites[n] )
			delete sprites[n];
	sprites.clear();
}

void CSceneMgr::Uninit()
{
	ReleaseDeviceObjects();

	gfx->UnregisterDeviceUser(this);

	// Tell the graphics object to drop the effects
	gfx->DropEffect(fxScene);

	ClearScene();

	for( unsigned int n = 0; n < materials.size(); n++ )
		if( materials[n] )
			delete materials[n];

	for( unsigned int n = 0; n < textures.size(); n++ )
		if( textures[n] )
			delete textures[n];
}

void CSceneMgr::Render()
{
	// Render
	if( initialized )
	{
		RenderScene();
	}
}

void CSceneMgr::SetCameraPosition(const vector2 &pos)
{
	cameraPos = pos;
}

vector2 CSceneMgr::GetCameraPosition()
{
	return cameraPos;
}

float CSceneMgr::GetCameraZoom()
{
	return cameraZoom;
}

void CSceneMgr::SetCameraZoom(float zoom)
{
	if( zoom < 0.0001f ) zoom = 0.0001f;
	else if( zoom > 1000.0f ) zoom = 1000.0f;
	cameraZoom = zoom;
}

void CSceneMgr::SetFixedViewport(bool v)
{
	fixedViewport = v;
}

vector2 CSceneMgr::ConvertScreenToWorld(int x, int y)
{
	// Transform the input coord to viewport coord
	int w, h;
	gfx->GetViewport(0, 0, &w, &h, 0, 0);

	vector2 coord(float(x)/w, float(y)/h);

	// The screen coord is inverted from the view coord
	coord.y = 1 - coord.y;

	if( fixedViewport )
	{
		// Multiply with the view's size
		coord.x *= 30.0f/cameraZoom;
		coord.y *= 22.5f/cameraZoom;

		// Translate to the middle of the view
		coord -= vector2(15.0f, 11.25f)/cameraZoom;
	}
	else
	{
		coord.x *= 0.1f*w/cameraZoom;
		coord.y *= 0.1f*h/cameraZoom;
		coord -= vector2(w*0.05f, h*0.05f)/cameraZoom;
	}

	// Translate with the camera position
	coord += cameraPos;

	return coord;
}

void CSceneMgr::PrepareRender()
{
	// Render the game objects
	gfx->SetEffect(fxScene);

	// Setup orthogonal view
	// Determine size of view port
	float minZ, maxZ;
	int w, h;
	gfx->GetViewport(0,0,&w,&h,&minZ,&maxZ);

	// Origin is in lower-left corner
	matrix4x4 mtx = matrix4x4::translation(-cameraPos.x, -cameraPos.y, 0);
	if( fixedViewport )
		mtx *= matrix4x4::orthoLH(30.0f/cameraZoom, 22.5f/cameraZoom, minZ, maxZ);
	else
		mtx *= matrix4x4::orthoLH(0.1f*w/cameraZoom, 0.1f*h/cameraZoom, minZ, maxZ);
	gfx->SetMatrix(MTX_VIEW, mtx);

	mtx.setIdentity();
	gfx->SetMatrix(MTX_PROJ, mtx);	
}

void CSceneMgr::RenderScene()
{
	numDrawCalls = 0;
	numMaterialChanges = 0;

	gfx->SetRenderTarget(0,0);

	gfx->Clear(CLEAR_TARGET|CLEAR_ZBUFFER, 0xFF777777, 1, 0);

	PrepareRender();

	// Visible screen
	float maxX, maxY;
	if( fixedViewport )
	{
		maxX = 15.0f/cameraZoom;
		maxY = 11.25f/cameraZoom;
	}
	else
	{
		int w, h;
		gfx->GetViewport(0,0,&w,&h,0,0);
		maxX = 0.1f*w/cameraZoom;
		maxY = 0.1f*h/cameraZoom;
	}

	// Only visible sprites should be considered
	visibleSprites.clear();
	for( unsigned int n = 0; n < sprites.size(); n++ )
	{
		CSprite *spr = sprites[n];
		if( spr->opaque == 0 ) continue;

		// Check if the sprite is within the currently visible area
		vector2 relPos = spr->pos - cameraPos;
		// Calculate the farthest away from the center the sprite can reach regardless of rotation
		float maxReach = sqrtf(spr->offset.lengthSq() + spr->size.lengthSq());
		if( fabs(relPos.x) - maxReach > maxX ) continue;
		if( fabs(relPos.y) - maxReach > maxY ) continue;

		visibleSprites.push_back(spr);
	}

	// Sort the visible sprites by their zOrder so that greater z is rendered first
	// Preserve the order of equal z's by using a stable sort algorithm
	struct compare 
	{ 
		bool operator() (CSprite *a, CSprite *b) 
		{ 
			if( a->zOrder > b->zOrder ) return true;
			return false;
		}
	} cmp;
	stable_sort(visibleSprites.begin(), visibleSprites.end(), cmp);

	// Render the sprites
	for( unsigned int n = 0; n < visibleSprites.size(); n++ )	
	{
		CSprite *sprite = visibleSprites[n];

		// TODO: Should sort by material, so it is not necessary to change the material as many times
		numMaterialChanges++;

		// TODO: Instead of setting the world matrix for each sprite, the draw  
		//       buffer should be set with the pre transformed coordinates
		matrix4x4 mtx;
		mtx = matrix4x4::translation(sprite->offset.x, sprite->offset.y, 0);
		mtx *= matrix4x4::rotationZ(sprite->rotation);
		mtx *= matrix4x4::translation(sprite->pos.x, sprite->pos.y, 0);
		gfx->SetMatrix(MTX_WORLD, mtx);

		CMaterial *mat = sprite->material < materials.size() ? materials[sprite->material] : 0;
		if( mat && mat->colorMap )
			gfx->SetTexture(mat->colorMap->tex, 0);
		else
			gfx->SetTexture(0, 0);

		gfx->CommitEffectChanges();

		unsigned int passes;
		gfx->BeginEffect(&passes);
		for( unsigned int p = 0; p < passes; p++ )
		{
			gfx->BeginEffectPass(p);

			DWORD color = mat ? mat->color : 0xFFFFFFFF;
			if( sprite->opaque != 1 )
			{
				DWORD alpha = DWORD((color>>24)*sprite->opaque);
				color = (color&0xFFFFFF) | (alpha<<24);
			}

			if( mat && mat->flipX )
			{
				// TODO: This should be optimized
				app->render->Begin(RENDER_QUAD_LIST);
				app->render->VtxColor(color);
				app->render->VtxTexCoord(1, 0);
				app->render->VtxPos(-0.5f*sprite->size.x, 0.5f*sprite->size.y, 0);
				app->render->VtxTexCoord(0, 0);
				app->render->VtxPos(0.5f*sprite->size.x, 0.5f*sprite->size.y, 0);
				app->render->VtxTexCoord(0, 1);
				app->render->VtxPos(0.5f*sprite->size.x, -0.5f*sprite->size.y, 0);
				app->render->VtxTexCoord(1, 1);
				app->render->VtxPos(-0.5f*sprite->size.x, -0.5f*sprite->size.y, 0);
				app->render->End();
				numDrawCalls++;
			}
			else
			{
				// TODO: This should be optimized
				app->render->Begin(RENDER_QUAD_LIST);
				app->render->VtxColor(color);
				app->render->VtxTexCoord(0, 0);
				app->render->VtxPos(-0.5f*sprite->size.x, 0.5f*sprite->size.y, 0);
				app->render->VtxTexCoord(1, 0);
				app->render->VtxPos(0.5f*sprite->size.x, 0.5f*sprite->size.y, 0);
				app->render->VtxTexCoord(1, 1);
				app->render->VtxPos(0.5f*sprite->size.x, -0.5f*sprite->size.y, 0);
				app->render->VtxTexCoord(0, 1);
				app->render->VtxPos(-0.5f*sprite->size.x, -0.5f*sprite->size.y, 0);
				app->render->End();
				numDrawCalls++;
			}

			gfx->EndEffectPass();
		}
		gfx->EndEffect();
	}
}

void CSceneMgr::GetStats(int *numDraws, int *numMatChanges)
{
	*numDraws = numDrawCalls;
	*numMatChanges = numMaterialChanges;
}

void CSceneMgr::AddSprite(CSprite *sprite)
{
	sprites.push_back(sprite);
}

void CSceneMgr::RemoveSprite(CSprite *sprite)
{
	for( unsigned int n = 0; n < sprites.size(); n++ )
	{
		if( sprites[n] == sprite )
		{
			sprites.erase(sprites.begin()+n);
			break;
		}
	}
}

void CSceneMgr::RemoveAllSprites()
{
	sprites.clear();
}

int CSceneMgr::GetMaterialIndex(const char *materialName)
{
	// First determine if the material is already loaded
	for( unsigned int n = 0; n < materials.size(); n++ )
	{
		if( materials[n] && materials[n]->name == materialName )
			return n;
	}

	// Load the material
	CMaterial *mat = new CMaterial;
	mat->Load(materialName);

	materials.push_back(mat);

	return (int)materials.size() - 1;
}

CMaterial *CSceneMgr::GetMaterialByIndex(int idx)
{
	if( idx < 0 || idx >= (signed int)materials.size() )
		return 0;

	return materials[idx];
}

CSharedTexture *CSceneMgr::GetSharedTexture(const char *textureName)
{
	// First determine if the texture is already loaded
	for( unsigned int n = 0; n < textures.size(); n++ )
	{
		if( textures[n] && textures[n]->name == textureName )
			return textures[n];
	}

	// Load the texture
	CSharedTexture *st = new CSharedTexture;
	st->name = textureName;

	string p = app->fileMgr->GetPathForFile(textureName);
	app->gfx->CreateTextureFromFile(p.c_str(), &st->tex);

	textures.push_back(st);

	return st;
}



