#include "scriptmgr.h"
#include "acutil_log.h"
#include "add_on\scriptbuilder\scriptbuilder.h"
#include "add_on\scriptstdstring\scriptstdstring.h"
#include "add_on\scriptarray\scriptarray.h"
#include <assert.h>
#include "acutil_input.h"
#include "application.h"
#include "gamemgr.h"
#include "gameobjectlink.h"
#include <string>
#include "acutil_filemgr.h"
#include "scriptmathbinding.h"
#include "guimgr.h"
#include "guielement.h"
#include "scenemgr.h"
#include "console.h"
#include "generator.h"

#ifdef LoadImage
#undef LoadImage
#endif

using namespace std;
using namespace acUtility;

void DebugWrite(const string &s)
{
	if( app->console )
		app->console->WriteMultipleLines(s.c_str());
}

// A wrapper for the CGenerator::AddObject to convert the script array to std::vector
void Generator_AddObject(const string &type, float x, float y, int zOrder, CScriptArray *nvpairs, CGenerator *generator)
{
	vector<string> nv;
	if( nvpairs )
	{
		for( asUINT n = 0; n < nvpairs->GetSize(); n++ )
			nv.push_back(*reinterpret_cast<string*>(nvpairs->At(n)));

		nvpairs->Release();
	}

	generator->AddObject(type, x, y, zOrder, nv);
}

// A wrapper for CGame::FindObjectsWithinRange to convert the std::vector to CScriptArray
CScriptArray *Game_FindObjectsWithinRange(const acUtility::vector2 &center, float radius, CGameMgr *gameMgr)
{
	vector<CGameObject*> objects;
	gameMgr->FindObjectsWithinRange(center, radius, objects);

	// Create the script array and populate it with the handles to the found entities
	CScriptArray *arr = new CScriptArray(objects.size(), app->scriptMgr->arrayOfEntityType);
	for( unsigned int n = 0; n < objects.size(); n++ )
	{
		*reinterpret_cast<const CGameObjectLink**>(arr->At(n)) = objects[n]->m_link;
		objects[n]->m_link->AddRef();
	}

	return arr;
}

// Our include callback routine
int ScriptIncludeCallback(const char *include, const char *from, CScriptBuilder *builder, void *param)
{
	CFileMgr *fileMgr = (CFileMgr*)param;
	
	// Determine the file path and add it to the builder
	string path = fileMgr->GetPathForFile(include);
	if( builder->AddSectionFromFile(path.c_str()) < 0 )
		return -1;

	return 0;
}

CScriptMgr::CScriptMgr()
{
	engine = 0;

	memset(&gui, 0, sizeof(SGui));
	memset(&game, 0, sizeof(SGame));

	totalAllocatedMemory = 0;
	numAllocations = 0;
	numFrees = 0;
	numAllocationsLastFrame = 0;
}

void *ScriptAlloc(size_t quantity)
{
	app->scriptMgr->totalAllocatedMemory += quantity;
	app->scriptMgr->numAllocations++;

	return malloc(quantity);
}

void ScriptFree(void *memory)
{
	app->scriptMgr->numFrees++;

	free(memory);
}

bool CScriptMgr::IsEntityTypeId(int typeId)
{
	if( !(typeId & asTYPEID_APPOBJECT) ) return false;

	// TODO: Can cache the type id, so we don't need to get the object type
	asIObjectType *type = engine->GetObjectTypeById(typeId);
	if( type && string(type->GetName()) == "CEntity" )
		return true;

	return false;
}

int CScriptMgr::Init()
{
	// Setup our memory routines
	asSetGlobalMemoryFunctions(ScriptAlloc, ScriptFree);
	
	// Create the script engine and configure it.
	engine = asCreateScriptEngine(ANGELSCRIPT_VERSION);
	if( engine == 0 )
	{
		LOG(("Failed to create the script engine"));
		return -1;
	}

	// Set the message callback
	engine->SetMessageCallback(asMETHOD(CScriptMgr, MessageCallback), this, asCALL_THISCALL);

	engine->SetDefaultAccessMask(ALL);

	// Register the entity controller interface.
	// The interface doesn't have any methods, as the script is not required to implement all of them.
	int r = engine->RegisterInterface("IEntityController"); assert( r >= 0 );

	// Register the basic types
	RegisterStdString(engine);
	RegisterScriptArray(engine, false);
	RegisterStdStringUtils(engine);
	RegisterScriptMath(engine);
	RegisterScriptHandle(engine);
	vector2TypeId = engine->GetTypeIdByDecl("vector2");

	// Register the debug write function
	r = engine->RegisterGlobalFunction("void debugWrite(const string &in)", asFUNCTION(DebugWrite), asCALL_CDECL); assert( r >= 0 );

	// Register the game object link that the script controller will use to access the game objects
	// TODO: Need GetWorldPosition and GetWorldYaw that will return the absolute pos/yaw, i.e. the entity's pos/yaw + that of the parent 
	r = engine->RegisterObjectType("CEntity", 0, asOBJ_REF); assert( r >= 0 );
	r = engine->RegisterObjectBehaviour("CEntity", asBEHAVE_ADDREF, "void f()", asMETHOD(CGameObjectLink, AddRef), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectBehaviour("CEntity", asBEHAVE_RELEASE, "void f()", asMETHOD(CGameObjectLink, Release), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "vector2 get_position() const", asMETHOD(CGameObjectLink, GetPosition), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void set_position(const vector2 &in)", asMETHOD(CGameObjectLink, SetPosition), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "vector2 get_velocity() const", asMETHOD(CGameObjectLink, GetVelocity), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void set_velocity(const vector2 &in)", asMETHOD(CGameObjectLink, SetVelocity), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "float get_rotation() const", asMETHOD(CGameObjectLink, GetRotation), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void set_rotation(float)", asMETHOD(CGameObjectLink, SetRotation), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void MoveTo(const vector2 &in, float speed)", asMETHOD(CGameObjectLink, MoveTo), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "const string &get_name() const", asMETHOD(CGameObjectLink, GetName), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void set_name(const string &in)", asMETHOD(CGameObjectLink, SetName), asCALL_THISCALL); assert( r >= 0 );
	// TODO: Need one for double, string, vector2, and CEntity as well
	r = engine->RegisterObjectMethod("CEntity", "bool GetProperty(const string &in, int64 &out) const", asMETHOD(CGameObjectLink, GetProperty), asCALL_THISCALL); assert( r >= 0 );
	// TODO: Need a method for querying the number of connected entities
	r = engine->RegisterObjectMethod("CEntity", "void Kill()", asMETHOD(CGameObjectLink, Kill), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "const string &get_type() const", asMETHOD(CGameObjectLink, GetType), asCALL_THISCALL); assert( r >= 0 );
	// TODO: Animation should also be property with get/set
	r = engine->RegisterObjectMethod("CEntity", "void SetAnimation(const string &in)", asMETHOD(CGameObjectLink, SetAnimation), asCALL_THISCALL); assert( r >= 0 );
	// TODO: Should allow setting the shape itself
	r = engine->RegisterObjectMethod("CEntity", "void SetCollision(bool)", asMETHOD(CGameObjectLink, SetCollision), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void SetGravity(const vector2 &in)", asMETHOD(CGameObjectLink, SetGravity), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void ClearGravity()", asMETHOD(CGameObjectLink, ClearGravity), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void set_awake(bool)", asMETHOD(CGameObjectLink, SetAwake), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "bool get_awake() const", asMETHOD(CGameObjectLink, GetAwake), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "float GetTimer(int) const", asMETHOD(CGameObjectLink, GetTimer), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void SetTimer(float, int)", asMETHOD(CGameObjectLink, SetTimer), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void KillTimer(int)", asMETHOD(CGameObjectLink, KillTimer), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void Touch(const CEntity @+ other)", asMETHOD(CGameObjectLink, Touch), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void SetTransparency(float)", asMETHOD(CGameObjectLink, SetTransparency), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CEntity", "void PlaySound(const string &in, float volume = 1)", asMETHOD(CGameObjectLink, PlaySound), asCALL_THISCALL); assert( r >= 0 );
	arrayOfEntityType = engine->GetObjectTypeById(engine->GetTypeIdByDecl("array<const CEntity @>"));

	// Register the game manager
	r = engine->RegisterObjectType("CGame", 0, asOBJ_REF | asOBJ_NOHANDLE); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "float get_frameDeltaTime()", asMETHOD(CGameMgr, GetFrameDeltaTime), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "float GetRandomNumber()", asMETHOD(CGameMgr, GetRandomNumber), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "const CEntity @FindEntityByName(const string &in)", asMETHOD(CGameMgr, FindObjectByName), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "array<const CEntity @> @FindEntitiesWithinRange(const vector2 &in, float)", asFUNCTION(Game_FindObjectsWithinRange), asCALL_CDECL_OBJLAST); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "CEntity @SpawnEntity(const string &in, const vector2 &in, const vector2 &in)", asMETHOD(CGameMgr, SpawnObject), asCALL_THISCALL); assert( r >= 0 );
	// TODO: The query parameter should be an enum with the possible values
	r = engine->RegisterObjectMethod("CGame", "float RayCast(const CEntity @ignore, int query, const vector2 &in s, const vector2 &in e, const CEntity @&out firstHit, vector2 &out vel)", asMETHOD(CGameMgr, RayCast), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "void SetCameraPosition(const vector2 &in)", asMETHOD(CGameMgr, SetCameraPosition), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "vector2 GetCameraPosition()", asMETHOD(CGameMgr, GetCameraPosition), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "vector2 ScreenToWorld(const vector2 &in)", asMETHOD(CGameMgr, ScreenToWorld), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "bool GetProperty(const string &in, int64 &out)", asMETHOD(CGameMgr, GetProperty), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "bool SetProperty(const string &in, int64)", asMETHOD(CGameMgr, SetProperty), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "void SendMessage(const CEntity @to, uint message, ref @arg)", asMETHOD(CGameMgr, SendMessage), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "vector2 GetNearestPathNode(const vector2 &in, const vector2 &in)", asMETHOD(CGameMgr, GetNearestPathNode), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "void SetGravity(const vector2 &in)", asMETHOD(CGameMgr, SetGravity), asCALL_THISCALL); assert( r >= 0 );
	engine->SetDefaultAccessMask(GAME);
	r = engine->RegisterObjectMethod("CGame", "int LoadLevel(const string &in)", asMETHOD(CGameMgr, LoadLevel), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "int GenerateLevel(const string &in)", asMETHOD(CGameMgr, GenerateLevel), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "void EndLevel(int)", asMETHOD(CGameMgr, EndLevel), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "void EndGame()", asMETHOD(CGameMgr, EndGame), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "void ShowMenu(uint)", asMETHOD(CGameMgr, ShowMenu), asCALL_THISCALL); assert( r >= 0 );
	engine->SetDefaultAccessMask(MENU);
	r = engine->RegisterObjectMethod("CGame", "void RestartLevel()", asMETHOD(CGameMgr, RestartLevel), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "void StartNewGame()", asMETHOD(CGameMgr, StartGame), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "void ResumeGame()", asMETHOD(CGameMgr, ResumeGame), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "void QuitGame()", asMETHOD(CGameMgr, QuitGame), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGame", "bool get_isGameOn()", asMETHOD(CGameMgr, IsGameOn), asCALL_THISCALL); assert( r >= 0 );
	engine->SetDefaultAccessMask(ALL);
	r = engine->RegisterGlobalProperty("CGame game", app->gameMgr);

	// Register the map generator class
	// We don't bother with counting the reference for the generator type, as the script won't be able to store it anywhere anyway
	engine->SetDefaultAccessMask(GENERATE);
	r = engine->RegisterObjectType("CGenerator", 0, asOBJ_REF | asOBJ_NOCOUNT); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGenerator", "void SetMapSize(uint, uint)", asMETHOD(CGenerator, SetMapSize), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGenerator", "void FillRect(uint,uint,uint,uint,uint8)", asMETHOD(CGenerator, FillRect), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGenerator", "uint8 &Tile(uint,uint)", asMETHOD(CGenerator, At), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGenerator", "void AddObject(const string &in, float, float, int, array<string> @)", asFUNCTION(Generator_AddObject), asCALL_CDECL_OBJLAST); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGenerator", "float GetRandomNumber()", asMETHOD(CGenerator, GetRandomNbr), asCALL_THISCALL); assert( r >= 0 );
	engine->SetDefaultAccessMask(ALL);

	// Register the input manager
	r = engine->RegisterObjectType("CInput", 0, asOBJ_REF | asOBJ_NOHANDLE); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CInput", "bool GetActionState(uint action) const", asMETHOD(acUtility::CInput, GetActionState), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CInput", "bool DidActionStart(uint action) const", asMETHOD(acUtility::CInput, DidActionStart), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterGlobalProperty("CInput input", app->input);

	// Register the gui element type
	// TODO: generator and entities shouldn't have access to the gui
	r = engine->RegisterObjectType("CGuiElement", 0, asOBJ_REF); assert( r >= 0 );
	r = engine->RegisterObjectBehaviour("CGuiElement", asBEHAVE_ADDREF, "void f()", asMETHOD(CGuiElement, AddRef), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectBehaviour("CGuiElement", asBEHAVE_RELEASE, "void f()", asMETHOD(CGuiElement, Release), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "string text", asOFFSET(CGuiElement, text)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "vector2 position", asOFFSET(CGuiElement, position)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "vector2 center", asOFFSET(CGuiElement, center)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "int srcImageIndex", asOFFSET(CGuiElement, srcImageIndex)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "vector2 srcBottomLeft", asOFFSET(CGuiElement, srcBottomLeft)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "vector2 srcTopRight", asOFFSET(CGuiElement, srcTopRight)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "float rotation", asOFFSET(CGuiElement, rotation)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "int zPosition", asOFFSET(CGuiElement, zPosition)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "vector2 scale", asOFFSET(CGuiElement, scale)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "int fontIndex", asOFFSET(CGuiElement, fontIndex)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "uint align", asOFFSET(CGuiElement, align)); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "uint color", asOFFSET(CGuiElement, color)); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGuiElement", "void SetTimer(float, int)", asMETHOD(CGuiElement, SetTimer), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGuiElement", "void KillTimer(int)", asMETHOD(CGuiElement, KillTimer), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterFuncdef("void OnTimerHandler(int, CGuiElement @)"); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "OnTimerHandler @onTimer", offsetof(CGuiElement, onTimerHandler)); assert( r >= 0 );
	engine->SetDefaultAccessMask(MENU);
	r = engine->RegisterFuncdef("void OnClickHandler(CGuiElement @)"); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "OnClickHandler @onClick", offsetof(CGuiElement, onClickHandler)); assert( r >= 0 );
	r = engine->RegisterFuncdef("void OnMouseOverHandler(bool, CGuiElement @)"); assert( r >= 0 );
	r = engine->RegisterObjectProperty("CGuiElement", "OnMouseOverHandler @onMouseOver", offsetof(CGuiElement, onMouseOverHandler)); assert( r >= 0 );
	engine->SetDefaultAccessMask(ALL);

	// Register the gui manager
	r = engine->RegisterObjectType("CGui", 0, asOBJ_REF | asOBJ_NOHANDLE); assert( r >= 0 );
	engine->SetDefaultAccessMask(GAME);
	r = engine->RegisterObjectMethod("CGui", "int LoadScript(const string &in)", asMETHOD(CGuiMgr, LoadScript), asCALL_THISCALL); assert( r >= 0 );
	engine->SetDefaultAccessMask(ALL);
	r = engine->RegisterObjectMethod("CGui", "int LoadImage(const string &in)", asMETHOD(CGuiMgr, LoadImage), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGui", "int LoadFont(const string &in)", asMETHOD(CGuiMgr, LoadFont), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGui", "CGuiElement @CreateElement(const vector2 &in, const vector2 &in, CGuiElement @+ parent = null)", asMETHOD(CGuiMgr, CreateElement), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterObjectMethod("CGui", "void SetMouseElement(CGuiElement @+)", asMETHOD(CGuiMgr, SetMouseElement), asCALL_THISCALL); assert( r >= 0 );
	r = engine->RegisterGlobalProperty("CGui gui", app->guiMgr); assert( r >= 0 );

	return 0;
}

void CScriptMgr::Uninit()
{
	for( unsigned int n = 0; n < controllers.size(); n++ )
		delete controllers[n];
	controllers.clear();

	for( unsigned int n = 0; n < contextPool.size(); n++ )
		contextPool[n]->Release();
	contextPool.clear();

	if( engine )
	{
		engine->Release();
		engine = 0;
	}
}

CScriptMgr::~CScriptMgr()
{
	Uninit();
}

int CScriptMgr::CallGeneratorScript(const char *script, CGenerator *generator)
{
	// TODO: The generator script should have specific access mask
	int r = LoadScript(script, script, -1);
	if( r < 0 ) return -1;

	asIScriptModule *mod = engine->GetModule(script);
	asIScriptFunction *func = mod->GetFunctionByDecl("void Generate(CGenerator @)");
	if( func )
	{
		// TODO: Should pass the asIScriptFunction directly
		asIScriptContext *ctx = PrepareContext(func);
		ctx->SetArgObject(0, generator);
		r = ExecuteCall(ctx);
		ReturnContextToPool(ctx);
	}

	// Discard the module again, since it won't be needed anymore
	engine->DiscardModule(script);

	if( r != asEXECUTION_FINISHED ) 
		return -1;

	return 0;
}

SEntityController *CScriptMgr::LoadControllerScript(const char *script)
{
	// First check if the script has been loaded already
	asIScriptModule *mod = engine->GetModule(script);
	if( mod )
	{
		// Find the type id from the cache
		for( unsigned int n = 0; n < controllers.size(); n++ )
		{
			if( controllers[n]->module == script )
			{
				return controllers[n];
			}
		}
		return 0;
	}
	
	int r = LoadScript(script, script, ENTITY);
	if( r < 0 ) return 0;

	// Find the script class that implements the IEntityController interface
	SEntityController *ctrl = new SEntityController;
	ctrl->module          = script;
	ctrl->typeId          = 0;
	ctrl->factoryFunc     = 0;
	ctrl->onThinkMethod   = 0;
	ctrl->onMessageMethod = 0;

	mod = engine->GetModule(script);
	asIObjectType *type = 0;
	int tc = mod->GetObjectTypeCount();
	for( int n = 0; n < tc; n++ )
	{
		bool found = false;
		type = mod->GetObjectTypeByIndex(n);
		int ic = type->GetInterfaceCount();
		for( int i = 0; i < ic; i++ )
		{
			if( strcmp(type->GetInterface(i)->GetName(), "IEntityController") == 0 )
			{
				found = true;
				break;
			}
		}

		if( found == true )
		{
			ctrl->typeId = type->GetTypeId();
			break;
		}
	}

	if( ctrl->typeId == 0 )
	{
		LOG(("No class implement the IEntityController interface in the script '%s'", script));
		app->console->WriteLine("Failed to locate any class that implements IEntityController");
		delete ctrl;
		return 0;
	}

	// Find the factory function
	string s = string(type->GetName()) + "@ " + string(type->GetName()) + "(CEntity @)";
	ctrl->factoryFunc = type->GetFactoryByDecl(s.c_str());

	if( ctrl->factoryFunc == 0 )
	{
		LOG(("Can't find '%s'", s.c_str()));
		app->console->WriteLine("Couldn't find proper class constructor");
		delete ctrl;
		return 0;
	}

	// Find the optional event handlers
	ctrl->onThinkMethod     = type->GetMethodByDecl("void onThink()");
	ctrl->onCollideMethod   = type->GetMethodByDecl("void onCollide(const CEntity @object, const vector2 &in position, const vector2 &in normal, float appliedImpulse)");
	ctrl->onProximityMethod = type->GetMethodByDecl("void onProximity(const CEntity @object, bool isNearing)");
	ctrl->onTimerMethod     = type->GetMethodByDecl("void onTimer(int)");
	ctrl->onTouchMethod     = type->GetMethodByDecl("void onTouch(const CEntity @origin)");
	ctrl->onMessageMethod   = type->GetMethodByDecl("void onMessage(uint msgId, ref @arg)");

	controllers.push_back(ctrl);

	// TODO: Store the SEntityController as user data in the asIObjectType
	//       This will allow the script manager to access it quickly without
	//       the game manager having to pass it with every function call.
	return ctrl;
}

asIScriptObject *CScriptMgr::CreateScriptObject(SEntityController *type, CGameObjectLink *link)
{
	// Call the factory function to create the engine
	asIScriptObject *ctrl = 0;
	asIScriptContext *ctx = PrepareContext(type->factoryFunc);
	if( ctx )
	{
		*((CGameObjectLink**)ctx->GetAddressOfArg(0)) = link;
		link->AddRef();
		int r = ExecuteCall(ctx);
		if( r == asEXECUTION_FINISHED )
		{
			// Get the newly created object and return it
			ctrl = *((asIScriptObject**)ctx->GetAddressOfReturnValue());
			ctrl->AddRef();
		}
		ReturnContextToPool(ctx);
	}

	return ctrl;
}

void CScriptMgr::CallEntityOnThink(SEntityController *type, asIScriptObject *obj)
{
	assert( obj );

	if( type->onThinkMethod )
		CallMethod(obj, type->onThinkMethod);
}

void CScriptMgr::CallEntityOnTouch(SEntityController *type, asIScriptObject *obj, const CGameObjectLink *origin)
{
	assert( obj );

	if( type->onTouchMethod )
	{
		asIScriptContext *ctx = PrepareContext(type->onTouchMethod); 
		if( ctx )
		{
			int r = ctx->SetObject(obj); assert( r >= 0 );

			// TODO: Error handling
			if( origin )
			{
				// This will automatically increase the reference count
				r = ctx->SetArgObject(0, (void*)origin); assert( r >= 0 );
			}
			ExecuteCall(ctx);
			ReturnContextToPool(ctx);
		}
	}
}

void CScriptMgr::CallEntityOnMessage(SEntityController *type, asIScriptObject *obj, unsigned int msgId, CScriptHandle arg)
{
	assert( obj );

	if( type->onMessageMethod )
	{
		asIScriptContext *ctx = PrepareContext(type->onMessageMethod);
		if( ctx )
		{
			int r = ctx->SetObject(obj); assert( r >= 0 );

			// TODO: error handling
			r = ctx->SetArgDWord(0, msgId); assert( r >= 0 );
			r = ctx->SetArgObject(1, &arg); assert( r >= 0 );

			ExecuteCall(ctx);
			ReturnContextToPool(ctx);
		}
	}
}


void CScriptMgr::CallEntityOnCollide(SEntityController *type, asIScriptObject *obj, const CGameObjectLink *obj2, const acUtility::vector2 &position, const acUtility::vector2 &normal, float appliedImpulse)
{
	assert( obj );

	if( type->onCollideMethod )
	{
		asIScriptContext *ctx = PrepareContext(type->onCollideMethod);
		if( ctx )
		{
			int r = ctx->SetObject(obj); assert( r >= 0 );

			// TODO: Error handling
			if( obj2 )
			{
				// This will automatically increase the reference count
				r = ctx->SetArgObject(0, (void*)obj2); assert( r >= 0 );
			}
			r = ctx->SetArgObject(1, (void*)&position); assert( r >= 0 );
			r = ctx->SetArgObject(2, (void*)&normal); assert( r >= 0 );
			r = ctx->SetArgFloat(3, appliedImpulse); assert( r >= 0 );
			ExecuteCall(ctx);
			ReturnContextToPool(ctx);
		}
	}
}

void CScriptMgr::CallEntityOnProximity(SEntityController *type, asIScriptObject *obj, const CGameObjectLink *obj2, bool isNearing)
{
	assert( obj );

	if( type->onProximityMethod )
	{
		asIScriptContext *ctx = PrepareContext(type->onProximityMethod);
		if( ctx )
		{
			int r = ctx->SetObject(obj); assert( r >= 0 );

			// TODO: Error handling
			if( obj2 )
			{
				// This will automatically increase the reference count
				r = ctx->SetArgObject(0, (void*)obj2); assert( r >= 0 );
			}
			r = ctx->SetArgByte(1, isNearing); assert( r >= 0 );
			ExecuteCall(ctx);
			ReturnContextToPool(ctx);
		}
	}
}

void CScriptMgr::CallEntityOnTimer(SEntityController *type, asIScriptObject *obj, int id)
{
	assert( obj );

	if( type->onTimerMethod )
	{
		asIScriptContext *ctx = PrepareContext(type->onTimerMethod);
		if( ctx )
		{
			int r = ctx->SetObject(obj); assert( r >= 0 );
			r = ctx->SetArgDWord(0, id); assert( r >= 0 );
			ExecuteCall(ctx);
			ReturnContextToPool(ctx);
		}
	}
}

// CallMethod can call either a global function or class method
void CScriptMgr::CallMethod(asIScriptObject *obj, asIScriptFunction *func)
{
	int r; 

	asIScriptContext *ctx = PrepareContext(func);
	if( ctx )
	{
		if( obj ) 
		{
			r = ctx->SetObject(obj); assert( r >= 0 );
		}
		ExecuteCall(ctx);
		ReturnContextToPool(ctx);
	}
}

void CScriptMgr::CallGuiOnClickHandler(asIScriptFunction *onClickHandler, CGuiElement *el)
{
	int r;
	asIScriptContext *ctx = PrepareContext(onClickHandler);
	if( ctx )
	{
		r = ctx->SetArgObject(0, el); assert( r >= 0 );
		ExecuteCall(ctx);
		ReturnContextToPool(ctx);
	}
}

void CScriptMgr::CallGuiOnMouseOverHandler(asIScriptFunction *onMouseOverHandler, bool isMouseOver, CGuiElement *el)
{
	int r;
	asIScriptContext *ctx = PrepareContext(onMouseOverHandler);
	if( ctx )
	{
		r = ctx->SetArgByte(0, isMouseOver); assert( r >= 0 );
		r = ctx->SetArgObject(1, el); assert( r >= 0 );
		ExecuteCall(ctx);
		ReturnContextToPool(ctx);
	}
}

void CScriptMgr::CallGuiOnTimerHandler(asIScriptFunction *onTimerHandler, int timerId, CGuiElement *el)
{
	int r;
	asIScriptContext *ctx = PrepareContext(onTimerHandler);
	if( ctx )
	{
		r = ctx->SetArgDWord(0, timerId); assert( r >= 0 );
		r = ctx->SetArgObject(1, el); assert( r >= 0 );
		ExecuteCall(ctx);
		ReturnContextToPool(ctx);
	}
}

asIScriptContext *CScriptMgr::PrepareContext(asIScriptFunction *func)
{
	asIScriptContext *ctx = 0;
	if( contextPool.size() )
	{
		ctx = *contextPool.rbegin();
		contextPool.pop_back();
	}
	else
		ctx = engine->CreateContext();

	// TODO: Error handling
	int r = ctx->Prepare(func); assert( r >= 0 );

	return ctx;
}

void CScriptMgr::ReturnContextToPool(asIScriptContext *ctx)
{
	contextPool.push_back(ctx);

	// Unprepare the context as we don't know when it will be used again
	ctx->Unprepare();
}

int CScriptMgr::ExecuteCall(asIScriptContext *ctx)
{
	int r = ctx->Execute();
	if( r != asEXECUTION_FINISHED )
	{
		LOG(("Script failed to execute (%d)", r));
		if( r == asEXECUTION_EXCEPTION )
		{
			LOG(("Exception: %s", ctx->GetExceptionString()));
			LOG(("Function: %s", ctx->GetExceptionFunction()->GetDeclaration()));
			LOG(("Line: %d", ctx->GetExceptionLineNumber()));

			char buf[256];
			sprintf_s(buf, 256, "Exception: %s", ctx->GetExceptionString());
			app->console->WriteLine(buf);

			sprintf_s(buf, 256, "Function: %s", ctx->GetExceptionFunction()->GetDeclaration());
			app->console->WriteLine(buf);

			sprintf_s(buf, 256, "Line: %d", ctx->GetExceptionLineNumber());
			app->console->WriteLine(buf);
		}
	}

	return r;
}

int CScriptMgr::LoadMenuScript(const char *script)
{
	int r = LoadScript(script, "menu", MENU);
	if( r < 0 ) return r;

	// Find the functions that the application may call
	asIScriptModule *mod = engine->GetModule("menu");
	menu.onShowMenu = mod->GetFunctionByDecl("void onShowMenu(uint)");

	// TODO: Should log if a function cannot be found so that the user knows why it is not called

	return 0;
}

int CScriptMgr::LoadGuiScript(const char *script)
{
	int r = LoadScript(script, "gui", GUI);
	if( r < 0 ) return r;

	// Find the functions that the application may call
	asIScriptModule *mod = engine->GetModule("gui");
	gui.onShowGui = mod->GetFunctionByDecl("void onShowGui()");
	gui.onFrame   = mod->GetFunctionByDecl("void onFrame()");

	// TODO: Should log if a function cannot be found so that the user knows why it is not called

	return 0;
}

int CScriptMgr::LoadGameScript(const char *script)
{
	int r = LoadScript(script, "game", GAME);
	if( r < 0 ) return r;

	// Find the functions that the application may call
	asIScriptModule *mod = engine->GetModule("game");
	game.onStartGame    = mod->GetFunctionByDecl("void onStartGame()");
	game.onEndLevel     = mod->GetFunctionByDecl("void onEndLevel(int)");
	game.onRestartLevel = mod->GetFunctionByDecl("void onRestartLevel()");

	return 0;
}

int CScriptMgr::LoadScript(const char *script, const char *moduleName, unsigned int accessMask)
{
	// Load and build the script
	messageBuffer = "";
	CScriptBuilder builder;
	builder.SetIncludeCallback(ScriptIncludeCallback, app->fileMgr);
	int r = builder.StartNewModule(engine, moduleName);
	if( r >= 0 )
		builder.GetModule()->SetAccessMask(accessMask);
	if( r >= 0 )
		r = builder.AddSectionFromFile(app->fileMgr->GetPathForFile(script).c_str());
	if( r >= 0 )
		r = builder.BuildModule();
	if( r < 0 )
	{
		LOG(("Failed to build script '%s' (%d)", script, r));
		app->console->WriteMultipleLines(messageBuffer.c_str());
		return -1;
	}
	else if( messageBuffer.length() )
	{
		app->console->WriteMultipleLines(messageBuffer.c_str());
	}

	return 0;
}

void CScriptMgr::CallMenuOnShowMenu(unsigned int id)
{
	if( menu.onShowMenu <= 0 )
		return;

	asIScriptContext *ctx = PrepareContext(menu.onShowMenu);
	if( ctx )
	{
		ctx->SetArgDWord(0, id);
		ExecuteCall(ctx);
		ReturnContextToPool(ctx);
	}
}

void CScriptMgr::CallGameOnStartGame()
{
	if( game.onStartGame <= 0 )
		return;

	CallMethod(0, game.onStartGame);
}

void CScriptMgr::CallGameOnRestartLevel()
{
	if( game.onRestartLevel <= 0 )
		return;

	CallMethod(0, game.onRestartLevel);
}

void CScriptMgr::CallGameOnEndLevel(int code)
{
	if( game.onEndLevel <= 0 )
		return;

	if( game.onEndLevel > 0 )
	{
		asIScriptContext *ctx = PrepareContext(game.onEndLevel);
		if( ctx )
		{
			int r = ctx->SetArgDWord(0, code); assert( r >= 0 );
			ExecuteCall(ctx);
			ReturnContextToPool(ctx);
		}
	}
}

void CScriptMgr::CallGuiOnShowGui()
{
	if( gui.onShowGui == 0 )
		return;

	CallMethod(0, gui.onShowGui);
}

void CScriptMgr::CallGuiOnFrame()
{
	if( gui.onFrame == 0 )
		return;

	CallMethod(0, gui.onFrame);
}

void CScriptMgr::MessageCallback(const asSMessageInfo *msg)
{
	const char *type = "ERR ";
	if( msg->type == asMSGTYPE_WARNING ) 
		type = "WARN";
	else if( msg->type == asMSGTYPE_INFORMATION ) 
		type = "INFO";

	char tmp[1024];
	sprintf_s(tmp, 1024, "%s (%d, %d) : %s : %s\n", msg->section, msg->row, msg->col, type, msg->message);
	messageBuffer += tmp;

	LOG(("%s (%d, %d) : %s : %s", msg->section, msg->row, msg->col, type, msg->message));
}

int CScriptMgr::GetGCObjects(UINT *totalDestroyed, UINT *totalDetected)
{
	// TODO: Get information on generations and on how many with circular references have been detected
	asUINT currentSize;
	engine->GetGCStatistics(&currentSize, totalDestroyed, totalDetected);
	return currentSize;
}

void CScriptMgr::GarbageCollect()
{
	// TODO: What algorithm do we need to use for garbage collection?
	// The script engine performs garbage collection, but we may need to perform additional passes
//	engine->GarbageCollect();
}

bool CScriptMgr::GetPropertyFromGameScript(const string &prop, __int64 &out)
{
	asIScriptModule *mod = engine->GetModule("game", asGM_ONLY_IF_EXISTS);
	if( mod )
	{
		int idx = mod->GetGlobalVarIndexByName(prop.c_str());
		if( idx >= 0 )
		{
			void *ptr = mod->GetAddressOfGlobalVar(idx);
			int typeId;
			mod->GetGlobalVar(idx, 0, 0, &typeId);

			return ConvertTypeToInt64(ptr, typeId, out);			
		}
	}

	return false;
}

bool CScriptMgr::SetPropertyFromGameScript(const string &prop, __int64 val)
{
	asIScriptModule *mod = engine->GetModule("game", asGM_ONLY_IF_EXISTS);
	if( !mod ) return false;

	int idx = mod->GetGlobalVarIndexByName(prop.c_str());
	if( idx >= 0 )
	{
		void *ptr = mod->GetAddressOfGlobalVar(idx);
		int typeId;
		bool isConst;
		mod->GetGlobalVar(idx, 0, 0, &typeId, &isConst);
		if( isConst ) return false;

		return ConvertInt64ToTypeAndAssign(ptr, typeId, val);
	}

	return false;
}

bool CScriptMgr::ConvertTypeToInt64(void *ptr, int typeId, __int64 &out)
{
	if( typeId == asTYPEID_BOOL )
		out = *(bool*)ptr ? 1 : 0;
	else if( typeId == asTYPEID_INT8 )
		out = *(signed char*)ptr;
	else if( typeId == asTYPEID_INT16 )
		out = *(signed short*)ptr;
	else if( typeId == asTYPEID_INT32 )
		out = *(int*)ptr;
	else if( typeId == asTYPEID_INT64 )
		out = *(__int64*)ptr;
	else if( typeId == asTYPEID_UINT8 )
		out = *(unsigned char*)ptr;
	else if( typeId == asTYPEID_UINT16 )
		out = *(unsigned short*)ptr;
	else if( typeId == asTYPEID_UINT32 )
		out = *(unsigned int*)ptr;
	else if( typeId == asTYPEID_UINT64 )
		out = *(unsigned __int64*)ptr;
	else if( typeId == asTYPEID_FLOAT )
		out = __int64(*(float*)ptr);
	else if( typeId == asTYPEID_DOUBLE )
		out = __int64(*(double*)ptr);
	else
		return false;

	return true;
}

bool CScriptMgr::ConvertInt64ToTypeAndAssign(void *ptr, int typeId, __int64 value)
{
	if( typeId == asTYPEID_BOOL )
		*(bool*)ptr = value ? true : false;
	else if( typeId == asTYPEID_INT8 )
		*(signed char*)ptr = (signed char)value;
	else if( typeId == asTYPEID_INT16 )
		*(signed short*)ptr = (signed short)value;
	else if( typeId == asTYPEID_INT32 )
		*(int*)ptr = (int)value;
	else if( typeId == asTYPEID_INT64 )
		*(__int64*)ptr = value;
	else if( typeId == asTYPEID_UINT8 )
		*(unsigned char*)ptr = (unsigned char)value;
	else if( typeId == asTYPEID_UINT16 )
		*(unsigned short*)ptr = (unsigned short)value;
	else if( typeId == asTYPEID_UINT32 )
		*(unsigned int*)ptr = (unsigned int)value;
	else if( typeId == asTYPEID_UINT64 )
		*(unsigned __int64*)ptr = (unsigned __int64)value;
	else if( typeId == asTYPEID_FLOAT )
		*(float*)ptr = (float)value;
	else if( typeId == asTYPEID_DOUBLE )
		*(double*)ptr = (double)value;
	else
		return false;

	return true;
}