#ifndef SCRIPTMGR_H
#define SCRIPTMGR_H

#include <angelscript.h>
#include <string>
#include <vector>
#include <acutil_math.h>

#include "add_on\scripthandle\scripthandle.h"

class CGameObjectLink;
class CGuiElement;
class CGenerator;

struct SEntityController
{
	std::string        module;
	int                typeId;
	asIScriptFunction *factoryFunc;
	asIScriptFunction *onThinkMethod;
	asIScriptFunction *onCollideMethod;
	asIScriptFunction *onProximityMethod;
	asIScriptFunction *onTimerMethod;
	asIScriptFunction *onTouchMethod;
	asIScriptFunction *onMessageMethod;
};

struct SGui
{
	asIScriptFunction *onShowGui;
	asIScriptFunction *onFrame;
};

struct SGame
{
	asIScriptFunction *onStartGame;
	asIScriptFunction *onEndLevel;
	asIScriptFunction *onRestartLevel;
};

struct SMenu
{
	asIScriptFunction *onShowMenu;
};

enum EACCESS
{
	GAME     = 1,
	ENTITY   = 2,
	MENU     = 4,
	GUI      = 8,
	GENERATE = 16,         // level generator
	ALL      = 0xFFFFFFFF
};

class CScriptMgr
{
public:
	CScriptMgr();
	~CScriptMgr();

	int Init();
	void Uninit();

	// Entity controller scripts
	// TODO: Must be possible to discard controller scripts, e.g. between levels
	SEntityController *LoadControllerScript(const char *script);
	asIScriptObject *CreateScriptObject(SEntityController *scriptType, CGameObjectLink *link);
	void CallEntityOnThink(SEntityController *scriptType, asIScriptObject *obj);
	void CallEntityOnCollide(SEntityController *scriptType, asIScriptObject *obj, const CGameObjectLink *obj2, const acUtility::vector2 &v, const acUtility::vector2 &n, float appliedImpulse);
	void CallEntityOnProximity(SEntityController *scriptType, asIScriptObject *obj, const CGameObjectLink *obj2, bool isNearing);
	void CallEntityOnTimer(SEntityController *scriptType, asIScriptObject *obj, int id);
	void CallEntityOnTouch(SEntityController *scriptType, asIScriptObject *obj, const CGameObjectLink *origin);
	void CallEntityOnMessage(SEntityController *scriptType, asIScriptObject *obj, unsigned int msgId, CScriptHandle arg);

	// Game script
	int LoadGameScript(const char *script);
	void CallGameOnStartGame();
	void CallGameOnEndLevel(int code);
	void CallGameOnRestartLevel();
	bool GetPropertyFromGameScript(const std::string &prop, __int64 &out);
	bool SetPropertyFromGameScript(const std::string &prop, __int64 value);

	// GUI script
	int LoadGuiScript(const char *script);
	void CallGuiOnShowGui();
	void CallGuiOnFrame();
	void CallGuiOnClickHandler(asIScriptFunction *onClickHandler, CGuiElement *el);
	void CallGuiOnMouseOverHandler(asIScriptFunction *onMouseOverHandler, bool isMouseOver, CGuiElement *el);
	void CallGuiOnTimerHandler(asIScriptFunction *onTimerHandler, int timerId, CGuiElement *el);

	// Generator script
	int CallGeneratorScript(const char *script, CGenerator *generator);

	// Menu script
	int LoadMenuScript(const char *script);
	void CallMenuOnShowMenu(unsigned int id);

	void GarbageCollect();

	// Helpers
	bool IsEntityTypeId(int typeId);
	bool ConvertTypeToInt64(void *ptr, int typeId, __int64 &out);
	bool ConvertInt64ToTypeAndAssign(void *ptr, int typeId, __int64 value);

	// Stats
	int GetGCObjects(asUINT *totalDestroyed, asUINT *totalDetected);

	size_t totalAllocatedMemory;
	size_t numAllocations;
	size_t numFrees;
	size_t numAllocationsLastFrame;

	// TODO: This doesn't need to be public
	int            vector2TypeId;
	asIObjectType *arrayOfEntityType;

protected:
	void MessageCallback(const asSMessageInfo *msg);
	void CallMethod(asIScriptObject *obj, asIScriptFunction *func);
	asIScriptContext *PrepareContext(asIScriptFunction *func);
	int  ExecuteCall(asIScriptContext *ctx);
	void ReturnContextToPool(asIScriptContext *ctx);
	int  LoadScript(const char *script, const char *moduleName, unsigned int accessMask);

	asIScriptEngine *engine;
	std::vector<SEntityController *> controllers; 
	SGui  gui;
	SGame game;
	SMenu menu;
	
	std::string messageBuffer;

	std::vector<asIScriptContext *> contextPool;
};

#endif